@@ -248,10 +248,10 @@ clutter_input_method_put_im_event (ClutterInputMethod      *im,
                                    uint32_t                 len,
                                    ClutterPreeditResetMode  mode)
 {
-  ClutterSeat *seat;
   ClutterEvent *event;
-
-  seat = clutter_backend_get_default_seat (clutter_get_default_backend ());
+  ClutterContext *context = _clutter_context_get_default ();
+  ClutterBackend *backend = clutter_context_get_backend (context);
+  ClutterSeat *seat = clutter_backend_get_default_seat (backend);
 
   event = clutter_event_im_new (event_type,
                                 CLUTTER_EVENT_FLAG_INPUT_METHOD,
@@ -522,6 +522,8 @@ clutter_input_method_forward_key (ClutterInputMethod *im,
   ClutterInputDevice *keyboard;
   ClutterSeat *seat;
   ClutterEvent *event;
+  ClutterContext *context;
+  ClutterBackend *backend;
 
   g_return_if_fail (CLUTTER_IS_INPUT_METHOD (im));
 
@@ -529,7 +531,9 @@ clutter_input_method_forward_key (ClutterInputMethod *im,
   if (!priv->focus)
     return;
 
-  seat = clutter_backend_get_default_seat (clutter_get_default_backend ());
+  context = _clutter_context_get_default ();
+  backend = clutter_context_get_backend (context);
+  seat = clutter_backend_get_default_seat (backend);
   keyboard = clutter_seat_get_keyboard (seat);
 
   event = clutter_event_key_new (press ? CLUTTER_KEY_PRESS : CLUTTER_KEY_RELEASE,
