@@ -18,15 +18,14 @@
 /**
  * ClutterPressGesture:
  *
- * A #ClutterGesture subclass for recognizing press gestures
+ * An abstract #ClutterGesture subclass building the base for recognizing press
+ * gestures
  */
 
-#include "config.h"
+#include "clutter/clutter-press-gesture.h"
 
-#include "clutter-press-gesture.h"
-
-#include "clutter-enum-types.h"
-#include "clutter-private.h"
+#include "clutter/clutter-enum-types.h"
+#include "clutter/clutter-private.h"
 
 typedef struct _ClutterPressGesturePrivate ClutterPressGesturePrivate;
 
@@ -36,7 +35,7 @@ struct _ClutterPressGesturePrivate
 
   int cancel_threshold;
 
-  unsigned int long_press_duration;
+  int long_press_duration_ms;
   unsigned int long_press_timeout_id;
 
   unsigned int n_presses_happened;
@@ -49,8 +48,6 @@ struct _ClutterPressGesturePrivate
   graphene_point_t press_coords;
   unsigned int press_button;
   ClutterModifierType modifier_state;
-
-  gboolean is_subclassed;
 };
 
 enum
@@ -58,31 +55,24 @@ enum
   PROP_0,
 
   PROP_CANCEL_THRESHOLD,
-  PROP_LONG_PRESS_DURATION,
+  PROP_LONG_PRESS_DURATION_MS,
   PROP_PRESSED,
   PROP_REQUIRED_BUTTON,
 
   PROP_LAST
 };
 
-enum
-{
-  PRESS,
-  LONG_PRESS,
-  RELEASE,
-
-  LAST_SIGNAL
-};
-
 static GParamSpec *obj_props[PROP_LAST] = { NULL, };
-static unsigned int obj_signals[LAST_SIGNAL] = { 0, };
 
-G_DEFINE_TYPE_WITH_PRIVATE (ClutterPressGesture, clutter_press_gesture, CLUTTER_TYPE_GESTURE)
+G_DEFINE_ABSTRACT_TYPE_WITH_PRIVATE (ClutterPressGesture,
+                                     clutter_press_gesture,
+                                     CLUTTER_TYPE_GESTURE)
 
 static unsigned int
 get_default_long_press_duration (void)
 {
-  ClutterSettings *settings = clutter_settings_get_default ();
+  ClutterContext *context = _clutter_context_get_default ();
+  ClutterSettings *settings = clutter_context_get_settings (context);
   int long_press_duration;
 
   g_object_get (settings,
@@ -95,7 +85,8 @@ get_default_long_press_duration (void)
 static unsigned int
 get_next_press_timeout_ms (void)
 {
-  ClutterSettings *settings = clutter_settings_get_default ();
+  ClutterContext *context = _clutter_context_get_default ();
+  ClutterSettings *settings = clutter_context_get_settings (context);
   int double_click_time;
 
   g_object_get (settings,
@@ -111,26 +102,17 @@ get_default_cancel_threshold (void)
   return 36;
 }
 
-static gboolean
+static void
 long_press_cb (gpointer user_data)
 {
   ClutterPressGesture *self = user_data;
   ClutterPressGesturePrivate *priv =
     clutter_press_gesture_get_instance_private (self);
 
-  if (priv->is_subclassed)
-    {
-      if (CLUTTER_PRESS_GESTURE_GET_CLASS (self)->long_press)
-        CLUTTER_PRESS_GESTURE_GET_CLASS (self)->long_press (self);
-    }
-  else
-    {
-      if (clutter_gesture_get_state (CLUTTER_GESTURE (self)) == CLUTTER_GESTURE_STATE_RECOGNIZING)
-        g_signal_emit (self, obj_signals[LONG_PRESS], 0);
-    }
+  if (CLUTTER_PRESS_GESTURE_GET_CLASS (self)->long_press)
+    CLUTTER_PRESS_GESTURE_GET_CLASS (self)->long_press (self);
 
   priv->long_press_timeout_id = 0;
-  return G_SOURCE_REMOVE;
 }
 
 static void
@@ -148,7 +130,7 @@ set_pressed (ClutterPressGesture *self,
   g_object_notify_by_pspec (G_OBJECT (self), obj_props[PROP_PRESSED]);
 }
 
-static gboolean
+static void
 next_press_timed_out (gpointer user_data)
 {
   ClutterPressGesture *self = user_data;
@@ -161,12 +143,11 @@ next_press_timed_out (gpointer user_data)
     clutter_gesture_set_state (gesture, CLUTTER_GESTURE_STATE_CANCELLED);
 
   priv->next_press_timeout_id = 0;
-  return G_SOURCE_REMOVE;
 }
 
 static gboolean
-should_handle_sequence (ClutterGesture     *gesture,
-                        const ClutterEvent *sequence_begin_event)
+clutter_press_gesture_should_handle_sequence (ClutterGesture     *gesture,
+                                              const ClutterEvent *sequence_begin_event)
 {
   ClutterEventType event_type = clutter_event_type (sequence_begin_event);
 
@@ -178,8 +159,8 @@ should_handle_sequence (ClutterGesture     *gesture,
 }
 
 static void
-point_began (ClutterGesture *gesture,
-             unsigned int    sequence)
+clutter_press_gesture_point_began (ClutterGesture *gesture,
+                                   unsigned int    sequence)
 {
   ClutterPressGesture *self = CLUTTER_PRESS_GESTURE (gesture);
   ClutterPressGesturePrivate *priv =
@@ -190,7 +171,7 @@ point_began (ClutterGesture *gesture,
   unsigned int press_button;
   ClutterModifierType modifier_state;
   graphene_point_t coords;
-  unsigned int long_press_duration;
+  unsigned int long_press_duration_ms;
 
   if (active_n_points != 1)
     {
@@ -222,13 +203,6 @@ point_began (ClutterGesture *gesture,
       priv->press_button = press_button;
       priv->modifier_state = modifier_state;
       priv->press_coords = coords;
-
-      if (!priv->is_subclassed)
-        {
-          clutter_gesture_set_state (gesture, CLUTTER_GESTURE_STATE_RECOGNIZING);
-          if (clutter_gesture_get_state (gesture) != CLUTTER_GESTURE_STATE_RECOGNIZING)
-            return;
-        }
     }
   else
     {
@@ -259,33 +233,25 @@ point_began (ClutterGesture *gesture,
     }
 
   priv->next_press_timeout_id =
-    g_timeout_add (get_next_press_timeout_ms (), next_press_timed_out, self);
+    g_timeout_add_once (get_next_press_timeout_ms (), next_press_timed_out, self);
 
-  long_press_duration = priv->long_press_duration == 0
-    ? get_default_long_press_duration()
-    : priv->long_press_duration;
+  long_press_duration_ms = priv->long_press_duration_ms == -1
+    ? get_default_long_press_duration ()
+    : priv->long_press_duration_ms;
 
   g_assert (priv->long_press_timeout_id == 0);
   priv->long_press_timeout_id =
-    g_timeout_add (long_press_duration, long_press_cb, self);
+    g_timeout_add_once (long_press_duration_ms, long_press_cb, self);
 
   set_pressed (self, TRUE);
 
-  if (priv->is_subclassed)
-    {
-      if (CLUTTER_PRESS_GESTURE_GET_CLASS (self)->press)
-        CLUTTER_PRESS_GESTURE_GET_CLASS (self)->press (self);
-    }
-  else
-    {
-      if (clutter_gesture_get_state (gesture) == CLUTTER_GESTURE_STATE_RECOGNIZING)
-        g_signal_emit (self, obj_signals[PRESS], 0);
-    }
+  if (CLUTTER_PRESS_GESTURE_GET_CLASS (self)->press)
+    CLUTTER_PRESS_GESTURE_GET_CLASS (self)->press (self);
 }
 
 static void
-point_moved (ClutterGesture *gesture,
-             unsigned int    sequence)
+clutter_press_gesture_point_moved (ClutterGesture *gesture,
+                                   unsigned int    sequence)
 {
   ClutterPressGesture *self = CLUTTER_PRESS_GESTURE (gesture);
   ClutterPressGesturePrivate *priv =
@@ -302,8 +268,8 @@ point_moved (ClutterGesture *gesture,
 }
 
 static void
-point_ended (ClutterGesture *gesture,
-             unsigned int    sequence)
+clutter_press_gesture_point_ended (ClutterGesture *gesture,
+                                   unsigned int    sequence)
 {
   ClutterPressGesture *self = CLUTTER_PRESS_GESTURE (gesture);
   ClutterPressGesturePrivate *priv =
@@ -318,6 +284,7 @@ point_ended (ClutterGesture *gesture,
    */
   event = clutter_gesture_get_point_event (gesture, sequence);
   modifier_state = clutter_event_get_state (event) &
+    CLUTTER_MODIFIER_MASK &
     ~(CLUTTER_BUTTON1_MASK | CLUTTER_BUTTON2_MASK | CLUTTER_BUTTON3_MASK |
       CLUTTER_BUTTON4_MASK | CLUTTER_BUTTON5_MASK);
 
@@ -329,18 +296,8 @@ point_ended (ClutterGesture *gesture,
   if (modifier_state != priv->modifier_state)
     priv->modifier_state = 0;
 
-  if (priv->is_subclassed)
-    {
-      if (CLUTTER_PRESS_GESTURE_GET_CLASS (self)->release)
-        CLUTTER_PRESS_GESTURE_GET_CLASS (self)->release (self);
-    }
-  else
-    {
-      if (clutter_gesture_get_state (gesture) == CLUTTER_GESTURE_STATE_RECOGNIZING)
-        g_signal_emit (self, obj_signals[RELEASE], 0);
-
-      clutter_gesture_set_state (gesture, CLUTTER_GESTURE_STATE_COMPLETED);
-    }
+  if (CLUTTER_PRESS_GESTURE_GET_CLASS (self)->release)
+    CLUTTER_PRESS_GESTURE_GET_CLASS (self)->release (self);
 
   set_pressed (self, FALSE);
 
@@ -354,13 +311,13 @@ point_ended (ClutterGesture *gesture,
 }
 
 static void
-crossing_event (ClutterGesture    *gesture,
-                unsigned int       point,
-                ClutterEventType   type,
-                uint32_t           time,
-                ClutterEventFlags  flags,
-                ClutterActor      *source_actor,
-                ClutterActor      *related_actor)
+clutter_press_gesture_crossing_event (ClutterGesture    *gesture,
+                                      unsigned int       point,
+                                      ClutterEventType   type,
+                                      uint32_t           time,
+                                      ClutterEventFlags  flags,
+                                      ClutterActor      *source_actor,
+                                      ClutterActor      *related_actor)
 {
   ClutterPressGesture *self = CLUTTER_PRESS_GESTURE (gesture);
   ClutterGestureState state = clutter_gesture_get_state (gesture);
@@ -371,9 +328,9 @@ crossing_event (ClutterGesture    *gesture,
 }
 
 static void
-state_changed (ClutterGesture      *gesture,
-               ClutterGestureState  old_state,
-               ClutterGestureState  new_state)
+clutter_press_gesture_state_changed (ClutterGesture      *gesture,
+                                     ClutterGestureState  old_state,
+                                     ClutterGestureState  new_state)
 {
   ClutterPressGesture *self = CLUTTER_PRESS_GESTURE (gesture);
   ClutterPressGesturePrivate *priv =
@@ -386,7 +343,10 @@ state_changed (ClutterGesture      *gesture,
 
       g_clear_handle_id (&priv->next_press_timeout_id, g_source_remove);
       g_clear_handle_id (&priv->long_press_timeout_id, g_source_remove);
+    }
 
+  if (new_state == CLUTTER_GESTURE_STATE_WAITING)
+    {
       priv->n_presses_happened = 0;
       priv->press_coords.x = 0;
       priv->press_coords.y = 0;
@@ -395,20 +355,6 @@ state_changed (ClutterGesture      *gesture,
     }
 }
 
-static void
-clutter_press_gesture_constructed (GObject *object)
-{
-  ClutterPressGesture *self = CLUTTER_PRESS_GESTURE (object);
-  ClutterPressGesturePrivate *priv =
-    clutter_press_gesture_get_instance_private (self);
-
-  G_OBJECT_CLASS (clutter_press_gesture_parent_class)->constructed (object);
-
-  priv->is_subclassed = CLUTTER_PRESS_GESTURE_GET_CLASS (self)->press != NULL ||
-    CLUTTER_PRESS_GESTURE_GET_CLASS (self)->release != NULL ||
-    CLUTTER_PRESS_GESTURE_GET_CLASS (self)->long_press != NULL;
-}
-
 static void
 clutter_press_gesture_set_property (GObject      *gobject,
                                     unsigned int  prop_id,
@@ -423,8 +369,8 @@ clutter_press_gesture_set_property (GObject      *gobject,
       clutter_press_gesture_set_cancel_threshold (self, g_value_get_int (value));
       break;
 
-    case PROP_LONG_PRESS_DURATION:
-      clutter_press_gesture_set_long_press_duration (self, g_value_get_uint (value));
+    case PROP_LONG_PRESS_DURATION_MS:
+      clutter_press_gesture_set_long_press_duration_ms (self, g_value_get_int (value));
       break;
 
     case PROP_REQUIRED_BUTTON:
@@ -451,8 +397,8 @@ clutter_press_gesture_get_property (GObject      *gobject,
       g_value_set_int (value, clutter_press_gesture_get_cancel_threshold (self));
       break;
 
-    case PROP_LONG_PRESS_DURATION:
-      g_value_set_uint (value, clutter_press_gesture_get_long_press_duration (self));
+    case PROP_LONG_PRESS_DURATION_MS:
+      g_value_set_int (value, clutter_press_gesture_get_long_press_duration_ms (self));
       break;
 
     case PROP_PRESSED:
@@ -475,16 +421,15 @@ clutter_press_gesture_class_init (ClutterPressGestureClass *klass)
   ClutterGestureClass *gesture_class = CLUTTER_GESTURE_CLASS (klass);
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
-  gobject_class->constructed = clutter_press_gesture_constructed;
   gobject_class->set_property = clutter_press_gesture_set_property;
   gobject_class->get_property = clutter_press_gesture_get_property;
 
-  gesture_class->should_handle_sequence = should_handle_sequence;
-  gesture_class->point_began = point_began;
-  gesture_class->point_moved = point_moved;
-  gesture_class->point_ended = point_ended;
-  gesture_class->crossing_event = crossing_event;
-  gesture_class->state_changed = state_changed;
+  gesture_class->should_handle_sequence = clutter_press_gesture_should_handle_sequence;
+  gesture_class->point_began = clutter_press_gesture_point_began;
+  gesture_class->point_moved = clutter_press_gesture_point_moved;
+  gesture_class->point_ended = clutter_press_gesture_point_ended;
+  gesture_class->crossing_event = clutter_press_gesture_crossing_event;
+  gesture_class->state_changed = clutter_press_gesture_state_changed;
 
   /**
    * ClutterPressGesture:cancel-threshold:
@@ -492,31 +437,27 @@ clutter_press_gesture_class_init (ClutterPressGestureClass *klass)
    * Threshold in pixels to cancel the gesture, use -1 to disable the threshold.
    */
   obj_props[PROP_CANCEL_THRESHOLD] =
-    g_param_spec_int ("cancel-threshold",
-                      "cancel-threshold",
-                      "cancel-threshold",
+    g_param_spec_int ("cancel-threshold", NULL, NULL,
                       -1, G_MAXINT, 0,
                       G_PARAM_READWRITE |
                       G_PARAM_STATIC_STRINGS |
                       G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * ClutterLongPressGesture:long-press-duration:
+   * ClutterPressGesture:long-press-duration-ms:
    *
    * The minimum duration of a press in milliseconds for it to be recognized
    * as a long press gesture.
    *
-   * A value of 0 (default) will make the gesture use the value of the
+   * A value of -1 (default) will make the gesture use the value of the
    * #ClutterSettings:long-press-duration property.
    */
-  obj_props[PROP_LONG_PRESS_DURATION] =
-    g_param_spec_uint ("long-press-duration",
-                       "long-press-duration",
-                       "long-press-duration",
-                       0, G_MAXUINT, 0,
-                       G_PARAM_READWRITE |
-                       G_PARAM_STATIC_STRINGS |
-                       G_PARAM_EXPLICIT_NOTIFY);
+  obj_props[PROP_LONG_PRESS_DURATION_MS] =
+    g_param_spec_int ("long-press-duration-ms", NULL, NULL,
+                      -1, G_MAXINT, -1,
+                      G_PARAM_READWRITE |
+                      G_PARAM_STATIC_STRINGS |
+                      G_PARAM_EXPLICIT_NOTIFY);
 
   /**
    * ClutterPressGesture:pressed:
@@ -524,9 +465,7 @@ clutter_press_gesture_class_init (ClutterPressGestureClass *klass)
    * Whether the clickable actor should be in "pressed" state
    */
   obj_props[PROP_PRESSED] =
-    g_param_spec_boolean ("pressed",
-                          "pressed",
-                          "pressed",
+    g_param_spec_boolean ("pressed", NULL, NULL,
                           FALSE,
                           G_PARAM_READWRITE |
                           G_PARAM_STATIC_STRINGS |
@@ -540,60 +479,13 @@ clutter_press_gesture_class_init (ClutterPressGestureClass *klass)
    * of the primary button.
    */
   obj_props[PROP_REQUIRED_BUTTON] =
-    g_param_spec_uint ("required-button",
-                       "required-button",
-                       "required-button",
+    g_param_spec_uint ("required-button", NULL, NULL,
                        0, G_MAXUINT, 0,
                        G_PARAM_READWRITE |
                        G_PARAM_STATIC_STRINGS |
                        G_PARAM_EXPLICIT_NOTIFY);
 
-  g_object_class_install_properties (gobject_class,
-                                     PROP_LAST,
-                                     obj_props);
-
-  /**
-   * ClutterPressGesture::press:
-   * @gesture: the #ClutterPressGesture that emitted the signal
-   *
-   * The ::press signal is emitted when a button is pressed or a touch point
-   * begins.
-   */
-  obj_signals[PRESS] =
-    g_signal_new ("press",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST,
-                  0,
-                  NULL, NULL, NULL,
-                  G_TYPE_NONE, 0, G_TYPE_NONE);
-
-  /**
-   * ClutterPressGesture::long-press:
-   * @gesture: the #ClutterPressGesture that emitted the signal
-   *
-   * The ::long-press gesture is emitted on long press.
-   */
-  obj_signals[LONG_PRESS] =
-    g_signal_new ("long-press",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST,
-                  0,
-                  NULL, NULL, NULL,
-                  G_TYPE_NONE, 0, G_TYPE_NONE);
-
-  /**
-   * ClutterPressGesture::release:
-   * @gesture: the #ClutterPressGesture that emitted the signal
-   *
-   * The ::release signal is emitted on button or touch release.
-   */
-  obj_signals[RELEASE] =
-    g_signal_new ("release",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST,
-                  0,
-                  NULL, NULL, NULL,
-                  G_TYPE_NONE, 0, G_TYPE_NONE);
+  g_object_class_install_properties (gobject_class, PROP_LAST, obj_props);
 }
 
 static void
@@ -605,7 +497,7 @@ clutter_press_gesture_init (ClutterPressGesture *self)
   priv->pressed = FALSE;
 
   priv->cancel_threshold = get_default_cancel_threshold ();
-  priv->long_press_duration = 0;
+  priv->long_press_duration_ms = -1;
 
   priv->n_presses_happened = 0;
   priv->next_press_timeout_id = 0;
@@ -617,19 +509,6 @@ clutter_press_gesture_init (ClutterPressGesture *self)
   priv->modifier_state = 0;
 }
 
-/**
- * clutter_press_gesture_new:
- *
- * Creates a new #ClutterPressGesture instance
- *
- * Returns: the newly created #ClutterPressGesture
- */
-ClutterAction *
-clutter_press_gesture_new (void)
-{
-  return g_object_new (CLUTTER_TYPE_PRESS_GESTURE, NULL);
-}
-
 /**
  * clutter_press_gesture_get_pressed:
  * @self: a #ClutterPressGesture
@@ -698,28 +577,29 @@ clutter_press_gesture_set_cancel_threshold (ClutterPressGesture *self,
 }
 
 /**
- * clutter_press_gesture_get_long_press_duration:
+ * clutter_press_gesture_get_long_press_duration_ms:
  * @self: a #ClutterPressGesture
  *
  * Gets the minimum duration is milliseconds that's necessary for a long press
- * to recognize.
+ * to recognize. A value of -1 means the default from
+ * #ClutterSettings:long-press-duration is used.
  */
-unsigned int
-clutter_press_gesture_get_long_press_duration (ClutterPressGesture *self)
+int
+clutter_press_gesture_get_long_press_duration_ms (ClutterPressGesture *self)
 {
   ClutterPressGesturePrivate *priv;
 
-  g_return_val_if_fail (CLUTTER_IS_PRESS_GESTURE (self), 0);
+  g_return_val_if_fail (CLUTTER_IS_PRESS_GESTURE (self), -1);
 
   priv = clutter_press_gesture_get_instance_private (self);
 
-  return priv->long_press_duration;
+  return priv->long_press_duration_ms;
 }
 
 /**
- * clutter_press_gesture_set_long_press_duration:
+ * clutter_press_gesture_set_long_press_duration_ms:
  * @self: a #ClutterPressGesture
- * @long_press_duration: minimum duration is ms for long press to recognize
+ * @long_press_duration_ms: minimum duration for long press to recognize
  *
  * Sets the minimum duration is milliseconds that's necessary for a long press
  * to recognize.
@@ -727,8 +607,8 @@ clutter_press_gesture_get_long_press_duration (ClutterPressGesture *self)
  * Pass -1 to use the default from #ClutterSettings:long-press-duration.
  */
 void
-clutter_press_gesture_set_long_press_duration (ClutterPressGesture *self,
-                                               unsigned int         long_press_duration)
+clutter_press_gesture_set_long_press_duration_ms (ClutterPressGesture *self,
+                                                  int                  long_press_duration_ms)
 {
   ClutterPressGesturePrivate *priv;
 
@@ -736,12 +616,12 @@ clutter_press_gesture_set_long_press_duration (ClutterPressGesture *self,
 
   priv = clutter_press_gesture_get_instance_private (self);
 
-  if (priv->long_press_duration == long_press_duration)
+  if (priv->long_press_duration_ms == long_press_duration_ms)
     return;
 
-  priv->long_press_duration = long_press_duration;
+  priv->long_press_duration_ms = long_press_duration_ms;
 
-  g_object_notify_by_pspec (G_OBJECT (self), obj_props[PROP_LONG_PRESS_DURATION]);
+  g_object_notify_by_pspec (G_OBJECT (self), obj_props[PROP_LONG_PRESS_DURATION_MS]);
 }
 
 /**
