@@ -30,16 +30,17 @@
 #include "core/boxes-private.h"
 
 #define SCALE_FACTORS_PER_INTEGER 4
-#define SCALE_FACTORS_STEPS (1.0 / (float) SCALE_FACTORS_PER_INTEGER)
+#define SCALE_FACTORS_STEPS (1.0f / (float) SCALE_FACTORS_PER_INTEGER)
 #define MINIMUM_SCALE_FACTOR 1.0f
 #define MAXIMUM_SCALE_FACTOR 4.0f
 #define MINIMUM_LOGICAL_AREA (560 * 320)
-#define MAXIMUM_REFRESH_RATE_DIFF 0.001
+#define MAXIMUM_REFRESH_RATE_DIFF 0.001f
 
 typedef struct _MetaMonitorMode
 {
   MetaMonitor *monitor;
   char *id;
+  unsigned int n_crtc_modes;
   MetaMonitorModeSpec spec;
   MetaMonitorCrtcMode *crtc_modes;
 } MetaMonitorMode;
@@ -66,18 +67,11 @@ typedef struct _MetaMonitorPrivate
 
   MetaLogicalMonitor *logical_monitor;
 
-  /*
-   * The primary or first output for this monitor, 0 if we can't figure out.
-   * It can be matched to a winsys_id of a MetaOutput.
-   *
-   * This is used as an opaque token on reconfiguration when switching from
-   * clone to extended, to decide on what output the windows should go next
-   * (it's an attempt to keep windows on the same monitor, and preferably on
-   * the primary one).
-   */
-  uint64_t winsys_id;
-
   char *display_name;
+
+  gboolean is_for_lease;
+
+  GList *color_modes;
 } MetaMonitorPrivate;
 
 G_DEFINE_TYPE_WITH_PRIVATE (MetaMonitor, meta_monitor, G_TYPE_OBJECT)
@@ -110,7 +104,7 @@ static void
 meta_monitor_mode_free (MetaMonitorMode *mode);
 
 MetaMonitorSpec *
-meta_monitor_spec_clone (MetaMonitorSpec *monitor_spec)
+meta_monitor_spec_clone (const MetaMonitorSpec *monitor_spec)
 {
   MetaMonitorSpec *new_monitor_spec;
 
@@ -137,8 +131,8 @@ meta_monitor_spec_hash (gconstpointer key)
 }
 
 gboolean
-meta_monitor_spec_equals (MetaMonitorSpec *monitor_spec,
-                          MetaMonitorSpec *other_monitor_spec)
+meta_monitor_spec_equals (const MetaMonitorSpec *monitor_spec,
+                          const MetaMonitorSpec *other_monitor_spec)
 {
   return (g_str_equal (monitor_spec->connector, other_monitor_spec->connector) &&
           g_str_equal (monitor_spec->vendor, other_monitor_spec->vendor) &&
@@ -211,6 +205,27 @@ meta_monitor_generate_spec (MetaMonitor *monitor)
   priv->spec = monitor_spec;
 }
 
+static void
+meta_monitor_init_supported_color_modes (MetaMonitor *monitor)
+{
+  MetaMonitorPrivate *priv = meta_monitor_get_instance_private (monitor);
+  const MetaOutputInfo *output_info =
+    meta_monitor_get_main_output_info (monitor);
+
+  priv->color_modes =
+    g_list_append (NULL, GINT_TO_POINTER (META_COLOR_MODE_DEFAULT));
+
+  if ((output_info->supported_color_spaces &
+       (1 << META_OUTPUT_COLORSPACE_BT2020)) &&
+      (output_info->supported_hdr_eotfs &
+       (1 << META_OUTPUT_HDR_METADATA_EOTF_PQ)))
+    {
+      priv->color_modes =
+        g_list_append (priv->color_modes,
+                       GINT_TO_POINTER (META_COLOR_MODE_BT2100));
+    }
+}
+
 static const double known_diagonals[] = {
     12.1,
     13.3,
@@ -424,20 +439,23 @@ meta_monitor_is_same_as (MetaMonitor *monitor,
 {
   const MetaMonitorSpec *spec = meta_monitor_get_spec (monitor);
   const MetaMonitorSpec *other_spec = meta_monitor_get_spec (other_monitor);
+  gboolean spec_is_unknown;
+  gboolean other_spec_is_unknown;
 
-  if ((g_strcmp0 (spec->vendor, "unknown") == 0 ||
-       g_strcmp0 (spec->product, "unknown") == 0 ||
-       g_strcmp0 (spec->serial, "unknown") == 0) &&
-      (g_strcmp0 (other_spec->vendor, "unknown") == 0 ||
-       g_strcmp0 (other_spec->product, "unknown") == 0 ||
-       g_strcmp0 (other_spec->serial, "unknown") == 0))
-    {
-      MetaMonitorPrivate *priv = meta_monitor_get_instance_private (monitor);
-      MetaMonitorPrivate *other_priv =
-        meta_monitor_get_instance_private (other_monitor);
+  spec_is_unknown =
+    g_strcmp0 (spec->vendor, "unknown") == 0 ||
+    g_strcmp0 (spec->product, "unknown") == 0 ||
+    g_strcmp0 (spec->serial, "unknown") == 0;
+  other_spec_is_unknown =
+    g_strcmp0 (other_spec->vendor, "unknown") == 0 ||
+    g_strcmp0 (other_spec->product, "unknown") == 0 ||
+    g_strcmp0 (other_spec->serial, "unknown") == 0;
 
-      return priv->winsys_id == other_priv->winsys_id;
-    }
+  if (spec_is_unknown && other_spec_is_unknown)
+    return g_strcmp0 (spec->connector, other_spec->connector) == 0;
+
+  if (spec_is_unknown || other_spec_is_unknown)
+    return FALSE;
 
   if (g_strcmp0 (spec->vendor, other_spec->vendor) != 0)
     return FALSE;
@@ -481,7 +499,7 @@ meta_monitor_get_physical_dimensions (MetaMonitor *monitor,
   *height_mm = output_info->height_mm;
 }
 
-CoglSubpixelOrder
+MetaSubpixelOrder
 meta_monitor_get_subpixel_order (MetaMonitor *monitor)
 {
   const MetaOutputInfo *output_info =
@@ -553,18 +571,18 @@ meta_monitor_get_connector_type (MetaMonitor *monitor)
   return output_info->connector_type;
 }
 
-MetaMonitorTransform
+MtkMonitorTransform
 meta_monitor_logical_to_crtc_transform (MetaMonitor          *monitor,
-                                        MetaMonitorTransform  transform)
+                                        MtkMonitorTransform  transform)
 {
   MetaOutput *output = meta_monitor_get_main_output (monitor);
 
   return meta_output_logical_to_crtc_transform (output, transform);
 }
 
-MetaMonitorTransform
-meta_monitor_crtc_to_logical_transform (MetaMonitor          *monitor,
-                                        MetaMonitorTransform  transform)
+MtkMonitorTransform
+meta_monitor_crtc_to_logical_transform (MetaMonitor         *monitor,
+                                        MtkMonitorTransform  transform)
 {
   MetaOutput *output = meta_monitor_get_main_output (monitor);
 
@@ -593,6 +611,7 @@ meta_monitor_finalize (GObject *object)
   MetaMonitor *monitor = META_MONITOR (object);
   MetaMonitorPrivate *priv = meta_monitor_get_instance_private (monitor);
 
+  g_list_free (priv->color_modes);
   g_hash_table_destroy (priv->mode_ids);
   g_list_free_full (priv->modes, (GDestroyNotify) meta_monitor_mode_free);
   meta_monitor_spec_free (priv->spec);
@@ -673,7 +692,7 @@ meta_monitor_create_spec (MetaMonitor  *monitor,
   const MetaCrtcModeInfo *crtc_mode_info =
     meta_crtc_mode_get_info (crtc_mode);
 
-  if (meta_monitor_transform_is_rotated (output_info->panel_orientation_transform))
+  if (mtk_monitor_transform_is_rotated (output_info->panel_orientation_transform))
     {
       int temp = width;
       width = height;
@@ -783,10 +802,11 @@ meta_monitor_normal_generate_modes (MetaMonitorNormal *monitor_normal)
                                              crtc_mode_info->height,
                                              crtc_mode);
       mode->id = generate_mode_id (&mode->spec);
+      mode->n_crtc_modes = 1;
       mode->crtc_modes = g_new (MetaMonitorCrtcMode, 1);
       mode->crtc_modes[0] = (MetaMonitorCrtcMode) {
         .output = output,
-        .crtc_mode = crtc_mode
+        .crtc_mode = g_object_ref (crtc_mode)
       };
 
       /*
@@ -840,8 +860,8 @@ meta_monitor_normal_new (MetaMonitorManager *monitor_manager,
   monitor_priv->outputs = g_list_append (NULL, g_object_ref (output));
   meta_output_set_monitor (output, monitor);
 
-  monitor_priv->winsys_id = meta_output_get_id (output);
   meta_monitor_generate_spec (monitor);
+  meta_monitor_init_supported_color_modes (monitor);
 
   meta_monitor_normal_generate_modes (monitor_normal);
 
@@ -902,12 +922,12 @@ meta_monitor_normal_get_suggested_position (MetaMonitor *monitor,
 }
 
 static void
-meta_monitor_normal_calculate_crtc_pos (MetaMonitor          *monitor,
-                                        MetaMonitorMode      *monitor_mode,
-                                        MetaOutput           *output,
-                                        MetaMonitorTransform  crtc_transform,
-                                        int                  *out_x,
-                                        int                  *out_y)
+meta_monitor_normal_calculate_crtc_pos (MetaMonitor         *monitor,
+                                        MetaMonitorMode     *monitor_mode,
+                                        MetaOutput          *output,
+                                        MtkMonitorTransform  crtc_transform,
+                                        int                 *out_x,
+                                        int                 *out_y)
 {
   *out_x = 0;
   *out_y = 0;
@@ -977,7 +997,7 @@ add_tiled_monitor_outputs (MetaGpu          *gpu,
 static void
 calculate_tile_coordinate (MetaMonitor         *monitor,
                            MetaOutput          *output,
-                           MetaMonitorTransform crtc_transform,
+                           MtkMonitorTransform  crtc_transform,
                            int                 *out_x,
                            int                 *out_y)
 {
@@ -994,8 +1014,8 @@ calculate_tile_coordinate (MetaMonitor         *monitor,
 
       switch (crtc_transform)
         {
-        case META_MONITOR_TRANSFORM_NORMAL:
-        case META_MONITOR_TRANSFORM_FLIPPED:
+        case MTK_MONITOR_TRANSFORM_NORMAL:
+        case MTK_MONITOR_TRANSFORM_FLIPPED:
           if ((other_output_info->tile_info.loc_v_tile ==
                output_info->tile_info.loc_v_tile) &&
               (other_output_info->tile_info.loc_h_tile <
@@ -1007,8 +1027,8 @@ calculate_tile_coordinate (MetaMonitor         *monitor,
                output_info->tile_info.loc_v_tile))
             y += other_output_info->tile_info.tile_h;
           break;
-        case META_MONITOR_TRANSFORM_180:
-        case META_MONITOR_TRANSFORM_FLIPPED_180:
+        case MTK_MONITOR_TRANSFORM_180:
+        case MTK_MONITOR_TRANSFORM_FLIPPED_180:
           if ((other_output_info->tile_info.loc_v_tile ==
                output_info->tile_info.loc_v_tile) &&
               (other_output_info->tile_info.loc_h_tile >
@@ -1020,8 +1040,8 @@ calculate_tile_coordinate (MetaMonitor         *monitor,
                output_info->tile_info.loc_v_tile))
             y += other_output_info->tile_info.tile_h;
           break;
-        case META_MONITOR_TRANSFORM_270:
-        case META_MONITOR_TRANSFORM_FLIPPED_270:
+        case MTK_MONITOR_TRANSFORM_270:
+        case MTK_MONITOR_TRANSFORM_FLIPPED_270:
           if ((other_output_info->tile_info.loc_v_tile ==
                output_info->tile_info.loc_v_tile) &&
               (other_output_info->tile_info.loc_h_tile >
@@ -1033,8 +1053,8 @@ calculate_tile_coordinate (MetaMonitor         *monitor,
                output_info->tile_info.loc_v_tile))
             x += other_output_info->tile_info.tile_h;
           break;
-        case META_MONITOR_TRANSFORM_90:
-        case META_MONITOR_TRANSFORM_FLIPPED_90:
+        case MTK_MONITOR_TRANSFORM_90:
+        case MTK_MONITOR_TRANSFORM_FLIPPED_90:
           if ((other_output_info->tile_info.loc_v_tile ==
                output_info->tile_info.loc_v_tile) &&
               (other_output_info->tile_info.loc_h_tile <
@@ -1182,8 +1202,9 @@ create_tiled_monitor_mode (MetaMonitorTiled *monitor_tiled,
     meta_monitor_create_spec (monitor, width, height, reference_crtc_mode);
   mode->parent.id = generate_mode_id (&mode->parent.spec);
 
+  mode->parent.n_crtc_modes = g_list_length (monitor_priv->outputs);
   mode->parent.crtc_modes = g_new0 (MetaMonitorCrtcMode,
-                                    g_list_length (monitor_priv->outputs));
+                                    mode->parent.n_crtc_modes);
   for (l = monitor_priv->outputs, i = 0; l; l = l->next, i++)
     {
       MetaOutput *output = l->data;
@@ -1200,7 +1221,7 @@ create_tiled_monitor_mode (MetaMonitorTiled *monitor_tiled,
 
       mode->parent.crtc_modes[i] = (MetaMonitorCrtcMode) {
         .output = output,
-        .crtc_mode = tiled_crtc_mode
+        .crtc_mode = g_object_ref (tiled_crtc_mode)
       };
 
       is_preferred = (is_preferred &&
@@ -1316,8 +1337,9 @@ create_untiled_monitor_mode (MetaMonitorTiled *monitor_tiled,
                                                 crtc_mode_info->height,
                                                 crtc_mode);
   mode->parent.id = generate_mode_id (&mode->parent.spec);
+  mode->parent.n_crtc_modes = g_list_length (monitor_priv->outputs);
   mode->parent.crtc_modes = g_new0 (MetaMonitorCrtcMode,
-                                    g_list_length (monitor_priv->outputs));
+                                    mode->parent.n_crtc_modes);
 
   for (l = monitor_priv->outputs, i = 0; l; l = l->next, i++)
     {
@@ -1327,7 +1349,7 @@ create_untiled_monitor_mode (MetaMonitorTiled *monitor_tiled,
         {
           mode->parent.crtc_modes[i] = (MetaMonitorCrtcMode) {
             .output = output,
-            .crtc_mode = crtc_mode
+            .crtc_mode = g_object_ref (crtc_mode)
           };
         }
       else
@@ -1553,7 +1575,6 @@ meta_monitor_tiled_new (MetaMonitorManager *monitor_manager,
   monitor_priv->backend = meta_monitor_manager_get_backend (monitor_manager);
 
   monitor_tiled->tile_group_id = output_info->tile_info.group_id;
-  monitor_priv->winsys_id = meta_output_get_id (output);
 
   monitor_tiled->origin_output = output;
   add_tiled_monitor_outputs (meta_output_get_gpu (output), monitor_tiled);
@@ -1617,10 +1638,10 @@ meta_monitor_tiled_derive_layout (MetaMonitor  *monitor,
     }
 
   *layout = (MtkRectangle) {
-    .x = roundf (min_x),
-    .y = roundf (min_y),
-    .width = roundf (max_x - min_x),
-    .height = roundf (max_y - min_y)
+    .x = (int) roundf (min_x),
+    .y = (int) roundf (min_y),
+    .width = (int) roundf (max_x - min_x),
+    .height = (int) roundf (max_y - min_y)
   };
 }
 
@@ -1633,12 +1654,12 @@ meta_monitor_tiled_get_suggested_position (MetaMonitor *monitor,
 }
 
 static void
-meta_monitor_tiled_calculate_crtc_pos (MetaMonitor          *monitor,
-                                       MetaMonitorMode      *monitor_mode,
-                                       MetaOutput           *output,
-                                       MetaMonitorTransform  crtc_transform,
-                                       int                  *out_x,
-                                       int                  *out_y)
+meta_monitor_tiled_calculate_crtc_pos (MetaMonitor         *monitor,
+                                       MetaMonitorMode     *monitor_mode,
+                                       MetaOutput          *output,
+                                       MtkMonitorTransform  crtc_transform,
+                                       int                 *out_x,
+                                       int                 *out_y)
 {
   MetaMonitorModeTiled *mode_tiled = (MetaMonitorModeTiled *) monitor_mode;
 
@@ -1688,6 +1709,8 @@ static void
 meta_monitor_mode_free (MetaMonitorMode *monitor_mode)
 {
   g_free (monitor_mode->id);
+  for (int i = 0; i < monitor_mode->n_crtc_modes; i++)
+    g_clear_object (&monitor_mode->crtc_modes[i].crtc_mode);
   g_free (monitor_mode->crtc_modes);
   g_free (monitor_mode);
 }
@@ -1721,12 +1744,12 @@ gboolean
 meta_monitor_mode_spec_has_similar_size (MetaMonitorModeSpec *monitor_mode_spec,
                                          MetaMonitorModeSpec *other_monitor_mode_spec)
 {
-  const float target_ratio = 1.0;
+  const float target_ratio = 1.0f;
   /* The a size difference of 15% means e.g. 4K modes matches other 4K modes,
    * FHD (2K) modes other FHD modes, and HD modes other HD modes, but not each
    * other.
    */
-  const float epsilon = 0.15;
+  const float epsilon = 0.15f;
 
   return G_APPROX_VALUE (((float) monitor_mode_spec->width /
                           other_monitor_mode_spec->width) *
@@ -1837,12 +1860,12 @@ meta_monitor_get_modes (MetaMonitor *monitor)
 }
 
 void
-meta_monitor_calculate_crtc_pos (MetaMonitor          *monitor,
-                                 MetaMonitorMode      *monitor_mode,
-                                 MetaOutput           *output,
-                                 MetaMonitorTransform  crtc_transform,
-                                 int                  *out_x,
-                                 int                  *out_y)
+meta_monitor_calculate_crtc_pos (MetaMonitor         *monitor,
+                                 MetaMonitorMode     *monitor_mode,
+                                 MetaOutput          *output,
+                                 MtkMonitorTransform  crtc_transform,
+                                 int                 *out_x,
+                                 int                 *out_y)
 {
   META_MONITOR_GET_CLASS (monitor)->calculate_crtc_pos (monitor,
                                                         monitor_mode,
@@ -1887,7 +1910,7 @@ calculate_scale (MetaMonitor                *monitor,
   int n_scales;
   float best_scale, best_dpi;
   int target_dpi;
-  const float scale_epsilon = 0.2;
+  const float scale_epsilon = 0.2f;
 
   /*
    * Somebody encoded the aspect ratio (16/9 or 16/10) instead of the physical
@@ -1900,7 +1923,7 @@ calculate_scale (MetaMonitor                *monitor,
   meta_monitor_get_physical_dimensions (monitor, &width_mm, &height_mm);
   if (width_mm == 0 || height_mm == 0)
     return 1.0;
-  diag_inches = sqrtf (width_mm * width_mm + height_mm * height_mm) / 25.4;
+  diag_inches = sqrtf (width_mm * width_mm + height_mm * height_mm) / 25.4f;
 
   /* Pick the appropriate target DPI based on screen size */
   if (diag_inches < UI_SCALE_LARGE_MIN_SIZE_INCHES)
@@ -1947,7 +1970,7 @@ calculate_scale (MetaMonitor                *monitor,
   if (constraints & META_MONITOR_SCALES_CONSTRAINT_NO_FRAC)
     {
       best_scale = floorf (MIN (scales[n_scales - 1],
-                                best_scale + 0.25 + scale_epsilon));
+                                best_scale + 0.25f + scale_epsilon));
     }
 
   return best_scale;
@@ -1984,8 +2007,8 @@ is_scale_valid_for_size (float width,
   if (scale < MINIMUM_SCALE_FACTOR || scale > MAXIMUM_SCALE_FACTOR)
     return FALSE;
 
-  return is_logical_size_large_enough (floorf (width / scale),
-                                       floorf (height / scale));
+  return is_logical_size_large_enough ((int) floorf (width / scale),
+                                       (int) floorf (height / scale));
 }
 
 gboolean
@@ -2004,11 +2027,11 @@ meta_monitor_mode_should_be_advertised (MetaMonitorMode *monitor_mode)
                                        monitor_mode->spec.height);
 }
 
-static float
-get_closest_scale_factor_for_resolution (float width,
-                                         float height,
-                                         float scale,
-                                         float threshold)
+float
+meta_get_closest_monitor_scale_factor_for_resolution (float width,
+                                                      float height,
+                                                      float scale,
+                                                      float threshold)
 {
   unsigned int i, j;
   float scaled_h;
@@ -2019,15 +2042,12 @@ get_closest_scale_factor_for_resolution (float width,
 
   best_scale = 0;
 
-  if (!is_scale_valid_for_size (width, height, scale))
-    return best_scale;
-
   if (fmodf (width, scale) == 0.0 && fmodf (height, scale) == 0.0)
     return scale;
 
   i = 0;
   found_one = FALSE;
-  base_scaled_w = floorf (width / scale);
+  base_scaled_w = (int) floorf (width / scale);
 
   do
     {
@@ -2105,9 +2125,13 @@ meta_monitor_calculate_supported_scales (MetaMonitor                 *monitor,
               float scale;
               float scale_value = i + j * SCALE_FACTORS_STEPS;
 
-              scale = get_closest_scale_factor_for_resolution (width, height,
-                                                               scale_value,
-                                                               max_bound);
+              if (!is_scale_valid_for_size (width, height, scale_value))
+                continue;
+
+              scale = meta_get_closest_monitor_scale_factor_for_resolution (width,
+                                                                            height,
+                                                                            scale_value,
+                                                                            max_bound);
               if (scale > 0.0)
                 g_array_append_val (supported_scales, scale);
             }
@@ -2286,66 +2310,30 @@ meta_monitor_get_min_refresh_rate (MetaMonitor *monitor,
                                                 min_refresh_rate);
 }
 
-gboolean
-meta_monitor_set_color_space (MetaMonitor           *monitor,
-                              MetaOutputColorspace   color_space,
-                              GError               **error)
+GList *
+meta_monitor_get_supported_color_modes (MetaMonitor *monitor)
 {
   MetaMonitorPrivate *priv = meta_monitor_get_instance_private (monitor);
-  GList *l;
 
-  for (l = priv->outputs; l; l = l->next)
-    {
-      MetaOutput *output = l->data;
-      const MetaOutputInfo *output_info = meta_output_get_info (output);
-
-      if (!(output_info->supported_color_spaces & (1 << color_space)))
-        {
-          g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
-                               "The color space is not supported by this monitor");
-          return FALSE;
-        }
-    }
-
-  for (l = priv->outputs; l; l = l->next)
-    {
-      MetaOutput *output = l->data;
-
-      meta_output_set_color_space (output, color_space);
-    }
-
-  return TRUE;
+  return priv->color_modes;
 }
 
 gboolean
-meta_monitor_set_hdr_metadata (MetaMonitor            *monitor,
-                               MetaOutputHdrMetadata  *metadata,
-                               GError                **error)
+meta_monitor_is_color_mode_supported (MetaMonitor   *monitor,
+                                      MetaColorMode  color_mode)
 {
   MetaMonitorPrivate *priv = meta_monitor_get_instance_private (monitor);
-  GList *l;
 
-  for (l = priv->outputs; l; l = l->next)
-    {
-      MetaOutput *output = l->data;
-      const MetaOutputInfo *output_info = meta_output_get_info (output);
+  return !!g_list_find (priv->color_modes, GINT_TO_POINTER (color_mode));
+}
 
-      if (!(output_info->supported_hdr_eotfs & (1 << metadata->eotf)))
-        {
-          g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
-                               "HDR metadata is not supported by this monitor");
-          return FALSE;
-        }
-    }
+MetaColorMode
+meta_monitor_get_color_mode (MetaMonitor *monitor)
+{
+  MetaOutput *output;
 
-  for (l = priv->outputs; l; l = l->next)
-    {
-      MetaOutput *output = l->data;
-
-      meta_output_set_hdr_metadata (output, metadata);
-    }
-
-  return TRUE;
+  output = meta_monitor_get_main_output (monitor);
+  return meta_output_get_color_mode (output);
 }
 
 gboolean
@@ -2381,8 +2369,8 @@ meta_parse_monitor_mode (const char *string,
     return FALSE;
   ptr++;
 
-  refresh_rate = g_ascii_strtod (ptr, &ptr);
-  if (refresh_rate == 0.0)
+  refresh_rate = (float) g_ascii_strtod (ptr, &ptr);
+  if (G_APPROX_VALUE (refresh_rate, 0.0f, FLT_EPSILON))
     return FALSE;
 
   if (ptr[0] != '\0')
@@ -2394,3 +2382,78 @@ out:
   *out_refresh_rate = refresh_rate;
   return TRUE;
 }
+
+gboolean
+meta_monitor_get_backlight_info (MetaMonitor *monitor,
+                                 int         *backlight_min,
+                                 int         *backlight_max)
+{
+  MetaOutput *main_output;
+  int value;
+
+  main_output = meta_monitor_get_main_output (monitor);
+  value = meta_output_get_backlight (main_output);
+  if (value >= 0)
+    {
+      const MetaOutputInfo *output_info = meta_output_get_info (main_output);
+      if (backlight_min)
+        *backlight_min = output_info->backlight_min;
+      if (backlight_max)
+        *backlight_max = output_info->backlight_max;
+      return TRUE;
+    }
+  else
+    {
+      return FALSE;
+    }
+}
+
+void
+meta_monitor_set_backlight (MetaMonitor *monitor,
+                            int          value)
+{
+  MetaMonitorPrivate *priv = meta_monitor_get_instance_private (monitor);
+  GList *l;
+
+  for (l = priv->outputs; l; l = l->next)
+    {
+      MetaOutput *output = l->data;
+
+      meta_output_set_backlight (output, value);
+    }
+}
+
+gboolean
+meta_monitor_get_backlight (MetaMonitor *monitor,
+                            int         *value)
+{
+  if (meta_monitor_get_backlight_info (monitor, NULL, NULL))
+    {
+      MetaOutput *main_output;
+
+      main_output = meta_monitor_get_main_output (monitor);
+      *value = meta_output_get_backlight (main_output);
+      return TRUE;
+    }
+  else
+    {
+      return FALSE;
+    }
+}
+
+void
+meta_monitor_set_for_lease (MetaMonitor *monitor,
+                            gboolean     for_lease)
+{
+  MetaMonitorPrivate *priv = meta_monitor_get_instance_private (monitor);
+
+  priv->is_for_lease = for_lease;
+}
+
+gboolean
+meta_monitor_is_for_lease (MetaMonitor *monitor)
+{
+  MetaMonitorPrivate *priv = meta_monitor_get_instance_private (monitor);
+
+  return priv->is_for_lease;
+}
