@@ -82,6 +82,9 @@ typedef struct _MetaWindowActorPrivate
 
   guint             updates_frozen         : 1;
   guint             first_frame_state      : 2; /* FirstFrameState */
+
+  /* whether the associated window was created during a window drag */
+  unsigned int tied_to_drag : 1;
 } MetaWindowActorPrivate;
 
 enum
@@ -375,6 +378,7 @@ meta_window_actor_real_assign_surface_actor (MetaWindowActor  *self,
   priv->surface = g_object_ref_sink (surface_actor);
 
   meta_window_actor_add_surface_actor (self, surface_actor);
+  meta_window_actor_update_opacity (self);
 
   if (meta_window_actor_is_frozen (self))
     meta_window_actor_set_frozen (self, TRUE);
@@ -495,7 +499,7 @@ init_surface_actor (MetaWindowActor *self)
   MetaWindow *window = priv->window;
   MetaSurfaceActor *surface_actor = NULL;
 
-#ifdef HAVE_X11_CLIENT
+#ifdef HAVE_X11
   if (!meta_is_wayland_compositor ())
     {
       surface_actor = meta_surface_actor_x11_new (window);
@@ -953,7 +957,7 @@ meta_window_actor_sync_actor_geometry (MetaWindowActor *self,
 
   /* When running as a Wayland compositor we catch size changes when new
    * buffers are attached */
-#ifdef HAVE_X11_CLIENT
+#ifdef HAVE_X11
   if (META_IS_SURFACE_ACTOR_X11 (priv->surface))
     meta_surface_actor_x11_set_size (META_SURFACE_ACTOR_X11 (priv->surface),
                                      actor_rect.width, actor_rect.height);
@@ -1201,8 +1205,8 @@ meta_window_actor_set_geometry_scale (MetaWindowActor *window_actor,
   priv->geometry_scale = geometry_scale;
 
   graphene_matrix_init_scale (&child_transform,
-                              1.0 / geometry_scale,
-                              1.0 / geometry_scale,
+                              1.f / geometry_scale,
+                              1.f / geometry_scale,
                               1);
   clutter_actor_set_child_transform (CLUTTER_ACTOR (window_actor),
                                      &child_transform);
@@ -1228,8 +1232,8 @@ meta_window_actor_get_buffer_bounds (MetaScreenCastWindow *screen_cast_window,
 
   stex = meta_surface_actor_get_texture (priv->surface);
   *bounds = (MtkRectangle) {
-    .width = floorf (meta_shaped_texture_get_unscaled_width (stex)),
-    .height = floorf (meta_shaped_texture_get_unscaled_height (stex)),
+    .width = (int) floorf (meta_shaped_texture_get_unscaled_width (stex)),
+    .height = (int) floorf (meta_shaped_texture_get_unscaled_height (stex)),
   };
 }
 
@@ -1268,9 +1272,8 @@ static gboolean
 meta_window_actor_transform_cursor_position (MetaScreenCastWindow *screen_cast_window,
                                              MetaCursorSprite     *cursor_sprite,
                                              graphene_point_t     *cursor_position,
-                                             float                *out_cursor_scale,
-                                             MetaMonitorTransform *out_cursor_transform,
-                                             graphene_point_t     *out_relative_cursor_position)
+                                             graphene_point_t     *out_relative_cursor_position,
+                                             float                *out_view_scale)
 {
   MetaWindowActor *window_actor = META_WINDOW_ACTOR (screen_cast_window);
   MetaWindowActorPrivate *priv =
@@ -1281,37 +1284,6 @@ meta_window_actor_transform_cursor_position (MetaScreenCastWindow *screen_cast_w
   if (!meta_window_has_pointer (window))
     return FALSE;
 
-  if (cursor_sprite &&
-      meta_cursor_sprite_get_cogl_texture (cursor_sprite) &&
-      out_cursor_scale)
-    {
-      MetaDisplay *display = meta_compositor_get_display (priv->compositor);
-      MetaContext *context = meta_display_get_context (display);
-      MetaBackend *backend = meta_context_get_backend (context);
-      MetaLogicalMonitor *logical_monitor;
-      float view_scale;
-      float cursor_texture_scale;
-
-      logical_monitor = meta_window_get_main_logical_monitor (window);
-
-      if (meta_backend_is_stage_views_scaled (backend))
-        view_scale = meta_logical_monitor_get_scale (logical_monitor);
-      else
-        view_scale = 1.0;
-
-      cursor_texture_scale = meta_cursor_sprite_get_texture_scale (cursor_sprite);
-
-      *out_cursor_scale = view_scale * cursor_texture_scale;
-    }
-
-  if (cursor_sprite &&
-      meta_cursor_sprite_get_cogl_texture (cursor_sprite) &&
-      out_cursor_transform)
-    {
-      *out_cursor_transform =
-        meta_cursor_sprite_get_texture_transform (cursor_sprite);
-    }
-
   if (out_relative_cursor_position)
     {
       MetaShapedTexture *stex = meta_surface_actor_get_texture (priv->surface);
@@ -1334,6 +1306,25 @@ meta_window_actor_transform_cursor_position (MetaScreenCastWindow *screen_cast_w
         out_relative_cursor_position->y *= unscaled_height / height;
     }
 
+  if (out_view_scale)
+    {
+      MetaDisplay *display = meta_compositor_get_display (priv->compositor);
+      MetaContext *context = meta_display_get_context (display);
+      MetaBackend *backend = meta_context_get_backend (context);
+
+      if (meta_backend_is_stage_views_scaled (backend))
+        {
+          MetaLogicalMonitor *logical_monitor;
+
+          logical_monitor = meta_window_get_main_logical_monitor (window);
+          *out_view_scale = meta_logical_monitor_get_scale (logical_monitor);
+        }
+      else
+        {
+          *out_view_scale = 1.0;
+        }
+    }
+
   return TRUE;
 }
 
@@ -1469,7 +1460,8 @@ meta_window_actor_blit_to_framebuffer (MetaScreenCastWindow *screen_cast_window,
 
   paint_context =
     clutter_paint_context_new_for_framebuffer (framebuffer, NULL,
-                                               CLUTTER_PAINT_FLAG_NONE);
+                                               CLUTTER_PAINT_FLAG_NONE,
+                                               clutter_actor_get_color_state (actor));
   clutter_actor_paint (actor, paint_context);
   clutter_paint_context_destroy (paint_context);
 
@@ -1576,12 +1568,12 @@ create_framebuffer_from_window_actor (MetaWindowActor  *self,
   resource_scale = clutter_actor_get_resource_scale (actor);
 
   texture = cogl_texture_2d_new_with_size (cogl_context,
-                                           clip->width * resource_scale,
-                                           clip->height * resource_scale);
+                                           (int) (clip->width * resource_scale),
+                                           (int) (clip->height * resource_scale));
   if (!texture)
     return NULL;
 
-  cogl_primitive_texture_set_auto_mipmap (texture, FALSE);
+  cogl_texture_2d_set_auto_mipmap (COGL_TEXTURE_2D (texture), FALSE);
 
   offscreen = cogl_offscreen_new_with_texture (texture);
   framebuffer = COGL_FRAMEBUFFER (offscreen);
@@ -1602,7 +1594,8 @@ create_framebuffer_from_window_actor (MetaWindowActor  *self,
 
   paint_context =
     clutter_paint_context_new_for_framebuffer (framebuffer, NULL,
-                                               CLUTTER_PAINT_FLAG_NONE);
+                                               CLUTTER_PAINT_FLAG_NONE,
+                                               clutter_actor_get_color_state (actor));
   clutter_actor_paint (actor, paint_context);
   clutter_paint_context_destroy (paint_context);
 
@@ -1677,10 +1670,10 @@ meta_window_actor_get_image (MetaWindowActor *self,
     goto out;
 
   framebuffer_clip = (MtkRectangle) {
-    .x = floorf (x),
-    .y = floorf (y),
-    .width = ceilf (width),
-    .height = ceilf (height),
+    .x = (int) floorf (x),
+    .y = (int) floorf (y),
+    .width = (int) ceilf (width),
+    .height = (int) ceilf (height),
   };
 
   if (clip)
@@ -1689,8 +1682,8 @@ meta_window_actor_get_image (MetaWindowActor *self,
       MtkRectangle intersected_clip;
 
       tmp_clip = *clip;
-      tmp_clip.x += floorf (x);
-      tmp_clip.y += floorf (y);
+      tmp_clip.x += (int) floorf (x);
+      tmp_clip.y += (int) floorf (y);
       if (!mtk_rectangle_intersect (&framebuffer_clip,
                                     &tmp_clip,
                                     &intersected_clip))
@@ -1707,14 +1700,14 @@ meta_window_actor_get_image (MetaWindowActor *self,
 
   resource_scale = clutter_actor_get_resource_scale (actor);
   surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
-                                        framebuffer_clip.width *
-                                        resource_scale,
-                                        framebuffer_clip.height *
-                                        resource_scale);
+                                        (int) (framebuffer_clip.width *
+                                               resource_scale),
+                                        (int) (framebuffer_clip.height *
+                                               resource_scale));
   cogl_framebuffer_read_pixels (framebuffer,
                                 0, 0,
-                                framebuffer_clip.width * resource_scale,
-                                framebuffer_clip.height * resource_scale,
+                                (int) (framebuffer_clip.width * resource_scale),
+                                (int) (framebuffer_clip.height * resource_scale),
                                 COGL_PIXEL_FORMAT_CAIRO_ARGB32_COMPAT,
                                 cairo_image_surface_get_data (surface));
 
@@ -1763,10 +1756,10 @@ meta_window_actor_paint_to_content (MetaWindowActor  *self,
     goto out;
 
   framebuffer_clip = (MtkRectangle) {
-    .x = floorf (x),
-    .y = floorf (y),
-    .width = ceilf (width),
-    .height = ceilf (height),
+    .x = (int) floorf (x),
+    .y = (int) floorf (y),
+    .width = (int) ceilf (width),
+    .height = (int) ceilf (height),
   };
 
   if (clip)
@@ -1794,3 +1787,20 @@ out:
   clutter_actor_uninhibit_culling (actor);
   return content;
 }
+
+void
+meta_window_actor_set_tied_to_drag (MetaWindowActor *window_actor,
+                                    gboolean         tied_to_drag)
+{
+  MetaWindowActorPrivate *priv =
+    meta_window_actor_get_instance_private (window_actor);
+  priv->tied_to_drag = tied_to_drag;
+}
+
+gboolean
+meta_window_actor_is_tied_to_drag (MetaWindowActor *window_actor)
+{
+  MetaWindowActorPrivate *priv =
+    meta_window_actor_get_instance_private (window_actor);
+  return priv->tied_to_drag;
+}
