@@ -34,14 +34,18 @@
 #include "backends/meta-logical-monitor.h"
 #include "clutter/clutter.h"
 #include "core/stack.h"
+#include "meta/meta-window-config.h"
 #include "meta/compositor.h"
 #include "meta/meta-close-dialog.h"
 #include "meta/util.h"
 #include "meta/window.h"
+#include "meta/meta-window-config.h"
 #include "wayland/meta-wayland-types.h"
 
 typedef struct _MetaWindowQueue MetaWindowQueue;
 
+#define META_WINDOW_TITLEBAR_HEIGHT 50
+
 typedef enum
 {
   META_CLIENT_TYPE_UNKNOWN = 0,
@@ -70,11 +74,17 @@ typedef enum
   META_MOVE_RESIZE_CONSTRAIN = 1 << 13,
 } MetaMoveResizeFlags;
 
+typedef enum _MetaPlaceFlag
+{
+  META_PLACE_FLAG_NONE = 0,
+  META_PLACE_FLAG_FORCE_MOVE = 1 << 0,
+  META_PLACE_FLAG_DENIED_FOCUS_AND_NOT_TRANSIENT = 1 << 1,
+} MetaPlaceFlag;
+
 typedef enum
 {
   META_MOVE_RESIZE_RESULT_MOVED               = 1 << 0,
   META_MOVE_RESIZE_RESULT_RESIZED             = 1 << 1,
-  META_MOVE_RESIZE_RESULT_FRAME_SHAPE_CHANGED = 1 << 2,
   META_MOVE_RESIZE_RESULT_STATE_CHANGED       = 1 << 3,
 } MetaMoveResizeResultFlags;
 
@@ -215,8 +225,6 @@ struct _MetaWindow
   MetaLogicalMonitor *highest_scale_monitor;
   MetaWorkspace *workspace;
   MetaWindowClientType client_type;
-  /* may be NULL! not all windows get decorated */
-  MetaFrame *frame;
   int depth;
   char *desc; /* used in debug spew */
   char *title;
@@ -263,7 +271,7 @@ struct _MetaWindow
 
   double tile_hfraction;
 
-  uint64_t preferred_output_winsys_id;
+  MetaLogicalMonitorId *preferred_logical_monitor;
 
   /* Area to cover when in fullscreen mode.  If _NET_WM_FULLSCREEN_MONITORS has
    * been overridden (via a client message), the window will cover the union of
@@ -276,9 +284,6 @@ struct _MetaWindow
     MetaLogicalMonitor *right;
   } fullscreen_monitors;
 
-  /* if non-NULL, the bounds of the window frame */
-  MtkRegion *frame_bounds;
-
   /* _NET_WM_WINDOW_OPACITY rescaled to 0xFF */
   guint8 opacity;
 
@@ -302,15 +307,14 @@ struct _MetaWindow
      know about for this window */
   guint32 net_wm_user_time;
 
-  gboolean has_custom_frame_extents;
   MetaFrameBorder custom_frame_extents;
 
   /* The rectangles here are in "frame rect" coordinates. See the
    * comment at the top of meta_window_move_resize_internal() for more
    * information. */
 
-  /* The current window geometry of the window. */
-  MtkRectangle rect;
+  /* The current configuration of the window. */
+  MetaWindowConfig *config;
 
   /* The geometry to restore when we unmaximize. */
   MtkRectangle saved_rect;
@@ -399,9 +403,6 @@ struct _MetaWindow
    * that to toggle between normal/tiled or maximized/tiled states. */
   guint saved_maximize : 1;
 
-  /* Whether we're fullscreen */
-  guint fullscreen : 1;
-
   /* Whether the window is marked as urgent */
   guint urgent : 1;
 
@@ -521,9 +522,6 @@ struct _MetaWindow
   /* Have we placed this window? */
   guint placed : 1;
 
-  /* Is this not a transient of the focus window which is being denied focus? */
-  guint denied_focus_and_not_transient : 1;
-
   /* Has this window not ever been shown yet? */
   guint showing_for_first_time : 1;
 
@@ -560,6 +558,9 @@ struct _MetaWindow
 
   guint in_workspace_change : 1;
   unsigned int can_grab : 1;
+
+  /* Whether the window surface is mapped. Only used on Wayland */
+  guint is_mapped : 1;
 };
 
 struct _MetaWindowClass
@@ -581,7 +582,6 @@ struct _MetaWindowClass
                                   MetaGrabOp  op);
   void (*current_workspace_changed) (MetaWindow *window);
   void (*move_resize_internal)   (MetaWindow                *window,
-                                  MetaGravity                gravity,
                                   MtkRectangle               unconstrained_rect,
                                   MtkRectangle               constrained_rect,
                                   MtkRectangle               temporary_rect,
@@ -620,32 +620,20 @@ struct _MetaWindowClass
   gboolean (*set_transient_for) (MetaWindow *window,
                                  MetaWindow *parent);
 
-  void (* map)   (MetaWindow *window);
-  void (* unmap) (MetaWindow *window);
-};
+  void (*stage_to_protocol) (MetaWindow *window,
+                             int         stage_x,
+                             int         stage_y,
+                             int        *protocol_x,
+                             int        *protocol_y);
+  void (*protocol_to_stage) (MetaWindow          *window,
+                             int                  protocol_x,
+                             int                  protocol_y,
+                             int                 *stage_x,
+                             int                 *stage_y,
+                             MtkRoundingStrategy  rounding_strategy);
 
-/* These differ from window->has_foo_func in that they consider
- * the dynamic window state such as "maximized", not just the
- * window's type
- */
-#define META_WINDOW_MAXIMIZED(w)       ((w)->maximized_horizontally && \
-                                        (w)->maximized_vertically)
-#define META_WINDOW_MAXIMIZED_VERTICALLY(w)    ((w)->maximized_vertically)
-#define META_WINDOW_MAXIMIZED_HORIZONTALLY(w)  ((w)->maximized_horizontally)
-#define META_WINDOW_TILED_SIDE_BY_SIDE(w)      ((w)->maximized_vertically && \
-                                                !(w)->maximized_horizontally && \
-                                                 (w)->tile_mode != META_TILE_NONE)
-#define META_WINDOW_TILED_LEFT(w)     (META_WINDOW_TILED_SIDE_BY_SIDE(w) && \
-                                       (w)->tile_mode == META_TILE_LEFT)
-#define META_WINDOW_TILED_RIGHT(w)    (META_WINDOW_TILED_SIDE_BY_SIDE(w) && \
-                                       (w)->tile_mode == META_TILE_RIGHT)
-#define META_WINDOW_TILED_MAXIMIZED(w)(META_WINDOW_MAXIMIZED(w) && \
-                                       (w)->tile_mode == META_TILE_MAXIMIZED)
-#define META_WINDOW_ALLOWS_MOVE(w)     ((w)->has_move_func && !(w)->fullscreen)
-#define META_WINDOW_ALLOWS_RESIZE_EXCEPT_HINTS(w)   ((w)->has_resize_func && !META_WINDOW_MAXIMIZED (w) && !(w)->fullscreen)
-#define META_WINDOW_ALLOWS_RESIZE(w)   (META_WINDOW_ALLOWS_RESIZE_EXCEPT_HINTS (w) &&                \
-                                        (((w)->size_hints.min_width < (w)->size_hints.max_width) ||  \
-                                         ((w)->size_hints.min_height < (w)->size_hints.max_height)))
+  MetaGravity (* get_gravity) (MetaWindow *window);
+};
 
 void        meta_window_unmanage           (MetaWindow  *window,
                                             guint32      timestamp);
@@ -677,11 +665,11 @@ gboolean    meta_window_has_fullscreen_monitors (MetaWindow *window);
 void        meta_window_adjust_fullscreen_monitor_rect (MetaWindow    *window,
                                                         MtkRectangle  *monitor_rect);
 
-void        meta_window_resize_frame_with_gravity (MetaWindow  *window,
-                                                   gboolean     user_op,
-                                                   int          w,
-                                                   int          h,
-                                                   MetaGravity  gravity);
+META_EXPORT_TEST
+void        meta_window_resize_frame (MetaWindow  *window,
+                                      gboolean     user_op,
+                                      int          w,
+                                      int          h);
 
 gboolean    meta_window_should_be_showing_on_workspace (MetaWindow    *window,
                                                         MetaWorkspace *workspace);
@@ -694,24 +682,6 @@ gboolean    meta_window_should_show (MetaWindow  *window);
 
 void        meta_window_update_struts      (MetaWindow  *window);
 
-/* gets position we need to set to stay in current position,
- * assuming position will be gravity-compensated. i.e.
- * this is the position a client would send in a configure
- * request.
- */
-void        meta_window_get_gravity_position (MetaWindow  *window,
-                                              MetaGravity  gravity,
-                                              int         *x,
-                                              int         *y);
-/* Get geometry for saving in the session; x/y are gravity
- * position, and w/h are in resize inc above the base size.
- */
-void        meta_window_get_session_geometry (MetaWindow  *window,
-                                              int         *x,
-                                              int         *y,
-                                              int         *width,
-                                              int         *height);
-
 gboolean    meta_window_geometry_contains_rect (MetaWindow   *window,
                                                 MtkRectangle *rect);
 
@@ -767,6 +737,9 @@ void meta_window_stack_just_below (MetaWindow *window,
 void meta_window_stack_just_above (MetaWindow *window,
                                    MetaWindow *above_this_one);
 
+int meta_window_stack_position_compare (gconstpointer window_a,
+                                        gconstpointer window_b);
+
 void meta_window_set_user_time (MetaWindow *window,
                                 guint32     timestamp);
 
@@ -805,10 +778,8 @@ void meta_window_set_opacity              (MetaWindow *window,
 gboolean meta_window_handle_ungrabbed_event (MetaWindow         *window,
                                              const ClutterEvent *event);
 
-void meta_window_get_client_area_rect (const MetaWindow *window,
-                                       MtkRectangle     *rect);
-void meta_window_get_titlebar_rect (MetaWindow   *window,
-                                    MtkRectangle *titlebar_rect);
+void meta_window_get_client_area_rect (MetaWindow   *window,
+                                       MtkRectangle *rect);
 
 void meta_window_activate_full (MetaWindow     *window,
                                 guint32         timestamp,
@@ -820,6 +791,7 @@ MetaLogicalMonitor * meta_window_find_monitor_from_frame_rect (MetaWindow *windo
 
 MetaLogicalMonitor * meta_window_find_monitor_from_id (MetaWindow *window);
 
+META_EXPORT_TEST
 MetaLogicalMonitor * meta_window_get_main_logical_monitor (MetaWindow *window);
 MetaLogicalMonitor * meta_window_get_highest_scale_monitor (MetaWindow *window);
 void meta_window_update_monitor (MetaWindow                   *window,
@@ -828,10 +800,9 @@ void meta_window_update_monitor (MetaWindow                   *window,
 void meta_window_set_urgent (MetaWindow *window,
                              gboolean    urgent);
 
-void meta_window_move_resize_internal (MetaWindow          *window,
-                                       MetaMoveResizeFlags  flags,
-                                       MetaGravity          gravity,
-                                       MtkRectangle         frame_rect);
+void meta_window_move_resize (MetaWindow          *window,
+                              MetaMoveResizeFlags  flags,
+                              MtkRectangle         frame_rect);
 
 void meta_window_grab_op_began (MetaWindow *window, MetaGrabOp op);
 void meta_window_grab_op_ended (MetaWindow *window, MetaGrabOp op);
@@ -845,10 +816,13 @@ void meta_window_ensure_close_dialog_timeout (MetaWindow *window);
 
 void meta_window_emit_size_changed (MetaWindow *window);
 
+void meta_window_emit_configure (MetaWindow       *window,
+                                 MetaWindowConfig *window_config);
+
 MetaPlacementRule *meta_window_get_placement_rule (MetaWindow *window);
 
-void meta_window_maybe_place (MetaWindow *window,
-                              gboolean    force_move);
+void meta_window_maybe_place (MetaWindow    *window,
+                              MetaPlaceFlag  flags);
 
 void meta_window_force_restore_shortcuts (MetaWindow         *window,
                                           ClutterInputDevice *source);
@@ -889,3 +863,45 @@ int meta_get_window_suspend_timeout_s (void);
 
 gboolean
 meta_window_should_attach_to_parent (MetaWindow *window);
+
+/**
+ * meta_window_set_normal_hints:
+ * @window:   The window to set the size hints on.
+ * @hints:    Either some size hints, or NULL for default.
+ *
+ * Sets the size hints for a window.  This happens when a
+ * WM_NORMAL_HINTS property is set on a window, but it is public
+ * because the size hints are set to defaults when a window is
+ * created.  See
+ * http://tronche.com/gui/x/icccm/sec-4.html#WM_NORMAL_HINTS
+ * for the X details.
+ */
+void meta_window_set_normal_hints (MetaWindow    *window,
+                                   MetaSizeHints *hints);
+
+void meta_window_stage_to_protocol_point (MetaWindow *window,
+                                          int         stage_x,
+                                          int         stage_y,
+                                          int        *protocol_x,
+                                          int        *protocol_y);
+
+void meta_window_protocol_to_stage_point (MetaWindow          *window,
+                                          int                  protocol_x,
+                                          int                  protocol_y,
+                                          int                 *stage_x,
+                                          int                 *stage_y,
+                                          MtkRoundingStrategy  rounding_strategy);
+
+gboolean meta_window_is_maximized (MetaWindow *window);
+
+gboolean meta_window_is_tiled_side_by_side (MetaWindow *window);
+
+gboolean meta_window_is_tiled_left (MetaWindow *window);
+
+gboolean meta_window_is_tiled_right (MetaWindow *window);
+
+MetaWindowConfig * meta_window_new_window_config (MetaWindow *window);
+
+MetaGravity meta_window_get_gravity (MetaWindow *window);
+
+void meta_window_set_mapped (MetaWindow *window, gboolean is_mapped);
