@@ -52,6 +52,48 @@ test_gesture_class_init (TestGestureClass *klass)
   gesture_class->point_ended = test_gesture_point_ended;
 }
 
+#define TEST_TYPE_RECOGNIZE_ON_PRESS_GESTURE test_recognize_on_press_gesture_get_type()
+
+static
+G_DECLARE_FINAL_TYPE (TestRecognizeOnPressGesture, test_recognize_on_press_gesture, TEST, RECOGNIZE_ON_PRESS_GESTURE, ClutterGesture)
+
+struct _TestRecognizeOnPressGesture
+{
+  ClutterGesture parent;
+};
+
+G_DEFINE_TYPE (TestRecognizeOnPressGesture, test_recognize_on_press_gesture, CLUTTER_TYPE_GESTURE);
+
+G_END_DECLS
+
+static gboolean
+test_recognize_on_press_gesture_should_handle_sequence (ClutterGesture     *self,
+                                                        const ClutterEvent *sequence_begin_event)
+{
+  return TRUE;
+}
+
+static void
+test_recognize_on_press_gesture_point_began (ClutterGesture *gesture,
+                                             unsigned int    sequence)
+{
+  clutter_gesture_set_state (gesture, CLUTTER_GESTURE_STATE_COMPLETED);
+}
+
+static void
+test_recognize_on_press_gesture_init (TestRecognizeOnPressGesture *self)
+{
+}
+
+static void
+test_recognize_on_press_gesture_class_init (TestRecognizeOnPressGestureClass *klass)
+{
+  ClutterGestureClass *gesture_class = CLUTTER_GESTURE_CLASS (klass);
+
+  gesture_class->should_handle_sequence = test_recognize_on_press_gesture_should_handle_sequence;
+  gesture_class->point_began = test_recognize_on_press_gesture_point_began;
+}
+
 static void
 gesture_changed_state_once (ClutterGesture      *gesture,
                             GParamSpec          *spec,
@@ -94,8 +136,10 @@ wait_stage_updated (gboolean *was_updated)
 static void
 gesture_relationship_freed_despite_relationship (void)
 {
-  ClutterAction *action_1 = CLUTTER_ACTION (g_object_new (TEST_TYPE_GESTURE, NULL));
-  ClutterAction *action_2 = CLUTTER_ACTION (g_object_new (TEST_TYPE_GESTURE, NULL));
+  ClutterAction *action_1, *action_2;
+
+  action_1 = g_object_new (TEST_TYPE_GESTURE, NULL);
+  action_2 = g_object_new (TEST_TYPE_GESTURE, NULL);
 
   g_object_add_weak_pointer (G_OBJECT (action_1), (gpointer *) &action_1);
   g_object_add_weak_pointer (G_OBJECT (action_2), (gpointer *) &action_2);
@@ -114,17 +158,22 @@ static void
 gesture_relationship_cancel_on_recognize (void)
 {
   ClutterActor *stage = clutter_test_get_stage ();
-  ClutterSeat *seat =
-    clutter_backend_get_default_seat (clutter_get_default_backend ());
+  ClutterSeat *seat = clutter_test_get_default_seat ();
   g_autoptr (ClutterVirtualInputDevice) virtual_pointer = NULL;
   int64_t now_us;
-  ClutterGesture *gesture_1 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-1", NULL));
-  ClutterGesture *gesture_2 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-2", NULL));
+  ClutterGesture *gesture_1, *gesture_2;
   gboolean was_updated;
 
   virtual_pointer = clutter_seat_create_virtual_device (seat, CLUTTER_POINTER_DEVICE);
   now_us = g_get_monotonic_time ();
 
+  gesture_1 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-1",
+                            NULL);
+  gesture_2 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-2",
+                            NULL);
+
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_1));
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_2));
 
@@ -173,18 +222,23 @@ static void
 gesture_relationship_simple (void)
 {
   ClutterActor *stage = clutter_test_get_stage ();
-  ClutterSeat *seat =
-    clutter_backend_get_default_seat (clutter_get_default_backend ());
+  ClutterSeat *seat = clutter_test_get_default_seat ();
   g_autoptr (ClutterVirtualInputDevice) virtual_pointer = NULL;
   int64_t now_us;
-  ClutterGesture *gesture_1 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-1", NULL));
-  ClutterGesture *gesture_2 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-2", NULL));
+  ClutterGesture *gesture_1, *gesture_2;
   gboolean was_updated;
   gboolean recognize_emitted, cancel_emitted;
 
   virtual_pointer = clutter_seat_create_virtual_device (seat, CLUTTER_POINTER_DEVICE);
   now_us = g_get_monotonic_time ();
 
+  gesture_1 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-1",
+                            NULL);
+  gesture_2 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-2",
+                            NULL);
+
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_1));
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_2));
 
@@ -235,17 +289,22 @@ static void
 gesture_relationship_two_points (void)
 {
   ClutterActor *stage = clutter_test_get_stage ();
-  ClutterSeat *seat =
-    clutter_backend_get_default_seat (clutter_get_default_backend ());
+  ClutterSeat *seat = clutter_test_get_default_seat ();
   g_autoptr (ClutterVirtualInputDevice) virtual_pointer = NULL;
   int64_t now_us;
-  ClutterGesture *gesture_1 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-1", NULL));
-  ClutterGesture *gesture_2 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-2", NULL));
+  ClutterGesture *gesture_1, *gesture_2;
   gboolean was_updated;
 
   virtual_pointer = clutter_seat_create_virtual_device (seat, CLUTTER_POINTER_DEVICE);
   now_us = g_get_monotonic_time ();
 
+  gesture_1 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-1",
+                            NULL);
+  gesture_2 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-2",
+                            NULL);
+
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_1));
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_2));
 
@@ -284,18 +343,25 @@ static void
 gesture_relationship_two_points_two_actors (void)
 {
   ClutterActor *stage = clutter_test_get_stage ();
-  ClutterSeat *seat =
-    clutter_backend_get_default_seat (clutter_get_default_backend ());
+  ClutterSeat *seat = clutter_test_get_default_seat ();
   g_autoptr (ClutterVirtualInputDevice) virtual_pointer = NULL;
   int64_t now_us;
-  ClutterActor *second_actor = clutter_actor_new ();
-  ClutterGesture *gesture_1 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-1", NULL));
-  ClutterGesture *gesture_2 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-2", NULL));
+  ClutterActor *second_actor;
+  ClutterGesture *gesture_1, *gesture_2;
   gboolean was_updated;
 
   virtual_pointer = clutter_seat_create_virtual_device (seat, CLUTTER_POINTER_DEVICE);
   now_us = g_get_monotonic_time ();
 
+  second_actor = clutter_actor_new ();
+
+  gesture_1 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-1",
+                            NULL);
+  gesture_2 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-2",
+                            NULL);
+
   clutter_actor_set_size (second_actor, 20, 20);
   clutter_actor_set_reactive (second_actor, TRUE);
   clutter_actor_add_child (stage, second_actor);
@@ -353,25 +419,27 @@ static void
 gesture_relationship_claim_new_sequence_while_already_recognizing (void)
 {
   ClutterActor *stage = clutter_test_get_stage ();
-  ClutterSeat *seat =
-    clutter_backend_get_default_seat (clutter_get_default_backend ());
+  ClutterSeat *seat = clutter_test_get_default_seat ();
   g_autoptr (ClutterVirtualInputDevice) virtual_pointer = NULL;
   int64_t now_us;
-  ClutterGesture *gesture_1 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-1", NULL));
-  ClutterGesture *gesture_2 =
-    CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE,
-                                   "name", "gesture-2",
-                                   NULL));
+  ClutterGesture *gesture_1, *gesture_2;
   gboolean was_updated;
   ClutterGestureState gesture_2_state_change;
 
+  virtual_pointer = clutter_seat_create_virtual_device (seat, CLUTTER_POINTER_DEVICE);
+  now_us = g_get_monotonic_time ();
+
+  gesture_1 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-1",
+                            NULL);
+  gesture_2 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-2",
+                            NULL);
+
   g_signal_connect (gesture_2, "notify::state",
                     G_CALLBACK (move_to_waiting_on_complete),
                     NULL);
 
-  virtual_pointer = clutter_seat_create_virtual_device (seat, CLUTTER_POINTER_DEVICE);
-  now_us = g_get_monotonic_time ();
-
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_1));
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_2));
 
@@ -451,18 +519,25 @@ static void
 gesture_relationship_claim_new_sequence_while_already_recognizing_2 (void)
 {
   ClutterActor *stage = clutter_test_get_stage ();
-  ClutterSeat *seat =
-    clutter_backend_get_default_seat (clutter_get_default_backend ());
+  ClutterSeat *seat = clutter_test_get_default_seat ();
   g_autoptr (ClutterVirtualInputDevice) virtual_pointer = NULL;
   int64_t now_us;
-  ClutterActor *second_actor = clutter_actor_new ();
-  ClutterGesture *gesture_1 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-1", NULL));
-  ClutterGesture *gesture_2 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-2", NULL));
+  ClutterActor *second_actor;
+  ClutterGesture *gesture_1, *gesture_2;
   gboolean was_updated;
 
   virtual_pointer = clutter_seat_create_virtual_device (seat, CLUTTER_POINTER_DEVICE);
   now_us = g_get_monotonic_time ();
 
+  second_actor = clutter_actor_new ();
+
+  gesture_1 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-1",
+                            NULL);
+  gesture_2 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-2",
+                            NULL);
+
   clutter_actor_add_action (stage, CLUTTER_ACTION (gesture_1));
 
   clutter_actor_set_size (second_actor, 20, 20);
@@ -534,15 +609,21 @@ gesture_relationship_unmap_before_points_added (void)
     clutter_backend_get_default_seat (clutter_get_default_backend ());
   g_autoptr (ClutterVirtualInputDevice) virtual_pointer = NULL;
   int64_t now_us;
-  ClutterActor *second_actor = clutter_actor_new ();
-  ClutterGesture *press_gesture = CLUTTER_GESTURE (g_object_new (CLUTTER_TYPE_PRESS_GESTURE, NULL));
-  ClutterGesture *gesture_2 = CLUTTER_GESTURE (g_object_new (TEST_TYPE_GESTURE, "name", "gesture-2", NULL));
+  ClutterActor *second_actor;
+  ClutterGesture *press_gesture, *gesture_2;
   gboolean was_updated;
   ClutterGestureState gesture_2_state_change;
 
   virtual_pointer = clutter_seat_create_virtual_device (seat, CLUTTER_POINTER_DEVICE);
   now_us = g_get_monotonic_time ();
 
+  second_actor = clutter_actor_new ();
+
+  press_gesture = g_object_new (TEST_TYPE_RECOGNIZE_ON_PRESS_GESTURE, NULL);
+  gesture_2 = g_object_new (TEST_TYPE_GESTURE,
+                            "name", "gesture-2",
+                            NULL);
+
   clutter_actor_add_action_full (stage, "press-gesture", CLUTTER_PHASE_CAPTURE, CLUTTER_ACTION (press_gesture));
 
   clutter_actor_set_size (second_actor, 20, 20);
@@ -565,7 +646,7 @@ gesture_relationship_unmap_before_points_added (void)
    * We test this to make sure nothing blows up: gesture_2 should move through
    * POSSIBLE into CANCELLED, then WAITING.
    */
-  g_signal_connect (press_gesture, "press", G_CALLBACK (unmap_second_actor_on_press),
+  g_signal_connect (press_gesture, "recognize", G_CALLBACK (unmap_second_actor_on_press),
                     second_actor);
 
   g_signal_connect (gesture_2, "notify::state",
@@ -575,7 +656,7 @@ gesture_relationship_unmap_before_points_added (void)
   clutter_virtual_input_device_notify_touch_down (virtual_pointer, now_us, 0, 15, 15);
   wait_stage_updated (&was_updated);
   g_assert_true (!clutter_actor_is_mapped (second_actor));
-  g_assert_true (clutter_gesture_get_state (press_gesture) == CLUTTER_GESTURE_STATE_RECOGNIZING);
+  g_assert_true (clutter_gesture_get_state (press_gesture) == CLUTTER_GESTURE_STATE_COMPLETED);
   g_assert_true (gesture_2_state_change == CLUTTER_GESTURE_STATE_POSSIBLE);
   g_assert_true (clutter_gesture_get_state (gesture_2) == CLUTTER_GESTURE_STATE_WAITING);
   g_assert_cmpint (clutter_gesture_get_n_points (press_gesture), ==, 1);
