@@ -36,13 +36,20 @@
 
 #include "backends/meta-dnd-private.h"
 #include "compositor/meta-dnd-actor-private.h"
+#include "compositor/meta-surface-actor.h"
+#include "compositor/meta-window-drag.h"
 #include "core/meta-selection-private.h"
+#include "meta/meta-debug.h"
 #include "meta/meta-selection-source-memory.h"
+#include "meta/meta-wayland-surface.h"
 #include "wayland/meta-selection-source-wayland-private.h"
+#include "wayland/meta-wayland-data-source.h"
 #include "wayland/meta-wayland-dnd-surface.h"
 #include "wayland/meta-wayland-pointer.h"
 #include "wayland/meta-wayland-private.h"
 #include "wayland/meta-wayland-seat.h"
+#include "wayland/meta-wayland-toplevel-drag.h"
+#include "wayland/meta-wayland-types.h"
 
 #define ROOTWINDOW_DROP_MIME "application/x-rootwindow-drop"
 
@@ -321,14 +328,6 @@ meta_wayland_drag_grab_get_focus (MetaWaylandDragGrab *drag_grab)
   return drag_grab->drag_focus;
 }
 
-void
-meta_wayland_drag_grab_update_feedback_actor (MetaWaylandDragGrab *drag_grab,
-                                              const ClutterEvent  *event)
-{
-  meta_feedback_actor_update (META_FEEDBACK_ACTOR (drag_grab->feedback_actor),
-                              event);
-}
-
 MetaWaylandSeat *
 meta_wayland_drag_grab_get_seat (MetaWaylandDragGrab *drag_grab)
 {
@@ -345,6 +344,18 @@ meta_wayland_drag_grab_get_device (MetaWaylandDragGrab    *drag_grab,
   return drag_grab->device;
 }
 
+MetaWaylandSurface *
+meta_wayland_drag_grab_get_origin (MetaWaylandDragGrab *drag_grab)
+{
+  return drag_grab->drag_origin;
+}
+
+MetaWaylandDataSource *
+meta_wayland_drag_grab_get_data_source (MetaWaylandDragGrab *drag_grab)
+{
+  return drag_grab->drag_data_source;
+}
+
 static void
 data_source_update_user_dnd_action (MetaWaylandDataSource *source,
                                     ClutterModifierType    modifiers)
@@ -481,6 +492,36 @@ drag_grab_focus (MetaWaylandEventHandler *handler,
     meta_wayland_drag_grab_set_focus (drag_grab, surface);
 }
 
+static void
+data_device_update_position (MetaWaylandDragGrab *drag_grab,
+                             graphene_point_t    *point)
+{
+  MetaWaylandCompositor *compositor =
+    meta_wayland_seat_get_compositor (drag_grab->seat);
+  MetaContext *context = meta_wayland_compositor_get_context (compositor);
+  MetaBackend *backend = meta_context_get_backend (context);
+  MetaMonitorManager *monitor_manager =
+    meta_backend_get_monitor_manager (backend);
+  MetaFeedbackActor *feedback_actor =
+    META_FEEDBACK_ACTOR (drag_grab->feedback_actor);
+  MetaLogicalMonitor *monitor;
+
+  if (!drag_grab->drag_surface)
+    return;
+
+  meta_feedback_actor_set_position (feedback_actor, point->x, point->y);
+
+  monitor = meta_monitor_manager_get_logical_monitor_at (monitor_manager,
+                                                         point->x, point->y);
+  meta_wayland_surface_set_main_monitor (drag_grab->drag_surface, monitor);
+}
+
+static gboolean
+is_dragging_window (MetaWaylandSeat *seat)
+{
+  return meta_wayland_data_device_get_toplevel_drag (&seat->data_device) != NULL;
+}
+
 static gboolean
 drag_grab_motion (MetaWaylandEventHandler *handler,
 		  const ClutterEvent      *event,
@@ -498,23 +539,20 @@ drag_grab_motion (MetaWaylandEventHandler *handler,
       drag_grab->sequence != clutter_event_get_event_sequence (event))
     return CLUTTER_EVENT_STOP;
 
+  clutter_event_get_position (event, &point);
+
   if (drag_grab->drag_focus)
     {
-      clutter_event_get_coords (event, &point.x, &point.y);
       time_ms = clutter_event_get_time (event);
       meta_wayland_surface_drag_dest_motion (drag_grab->drag_focus,
                                              point.x, point.y, time_ms);
     }
 
-  if (drag_grab->drag_surface)
-    {
-      meta_feedback_actor_update (META_FEEDBACK_ACTOR (drag_grab->feedback_actor),
-                                  event);
-    }
+  data_device_update_position (drag_grab, &point);
 
   meta_dnd_wayland_on_motion_event (meta_backend_get_dnd (backend), event);
 
-  return CLUTTER_EVENT_STOP;
+  return !is_dragging_window (drag_grab->seat);
 }
 
 static gboolean
@@ -525,6 +563,7 @@ drag_grab_release (MetaWaylandEventHandler *handler,
   MetaWaylandDragGrab *drag_grab = user_data;
   MetaWaylandSeat *seat = drag_grab->seat;
   MetaWaylandDataSource *source = drag_grab->drag_data_source;
+  MetaWaylandToplevelDrag *toplevel_drag;
   gboolean success;
 
   if (drag_grab->device != clutter_event_get_device (event) ||
@@ -539,12 +578,22 @@ drag_grab_release (MetaWaylandEventHandler *handler,
                            CLUTTER_BUTTON5_MASK)) > 1)
     return CLUTTER_EVENT_STOP;
 
+  toplevel_drag = meta_wayland_data_device_get_toplevel_drag (&seat->data_device);
+  if (toplevel_drag)
+    {
+      meta_topic (META_DEBUG_WAYLAND, "Will end xdg_toplevel_drag#%u.",
+                  wl_resource_get_id (toplevel_drag->resource));
+      meta_wayland_data_source_notify_drop_performed (source);
+      meta_wayland_toplevel_drag_end (toplevel_drag);
+    }
+
   if (drag_grab->drag_focus && source &&
       meta_wayland_data_source_has_target (source) &&
       meta_wayland_data_source_get_current_action (source))
     {
       meta_wayland_surface_drag_dest_drop (drag_grab->drag_focus);
-      meta_wayland_data_source_notify_drop_performed (source);
+      if (!meta_wayland_data_source_get_drop_performed (source))
+        meta_wayland_data_source_notify_drop_performed (source);
 
       meta_wayland_data_source_update_in_ask (source);
       success = TRUE;
@@ -583,11 +632,20 @@ drag_grab_key (MetaWaylandEventHandler *handler,
                const ClutterEvent      *event,
                gpointer                 user_data)
 {
+  MetaWaylandToplevelDrag *toplevel_drag;
   MetaWaylandDragGrab *drag_grab = user_data;
   ClutterModifierType modifiers;
 
   if (clutter_event_get_key_symbol (event) == CLUTTER_KEY_Escape)
     {
+      toplevel_drag = meta_wayland_data_device_get_toplevel_drag (&drag_grab->seat->data_device);
+      if (toplevel_drag)
+        {
+          meta_topic (META_DEBUG_WAYLAND, "Will cancel xdg_toplevel_drag#%u.",
+                      wl_resource_get_id (toplevel_drag->resource));
+          meta_wayland_toplevel_drag_end (toplevel_drag);
+        }
+
       meta_wayland_data_device_set_dnd_source (&drag_grab->seat->data_device,
                                                NULL);
       unset_selection_source (&drag_grab->seat->data_device, META_SELECTION_DND);
@@ -710,8 +768,8 @@ meta_wayland_data_device_start_drag (MetaWaylandDataDevice           *data_devic
                                        drag_start.x,
                                        drag_start.y,
                                        &surface_pos.x, &surface_pos.y);
-  drag_grab->drag_start_x = surface_pos.x;
-  drag_grab->drag_start_y = surface_pos.y;
+  drag_grab->drag_start_x = (int) surface_pos.x;
+  drag_grab->drag_start_y = (int) surface_pos.y;
 
   drag_grab->need_initial_focus = TRUE;
 
@@ -748,8 +806,7 @@ meta_wayland_data_device_start_drag (MetaWaylandDataDevice           *data_devic
                                       0, 0);
       clutter_actor_add_child (drag_grab->feedback_actor, drag_surface_actor);
 
-      meta_feedback_actor_set_position (META_FEEDBACK_ACTOR (drag_grab->feedback_actor),
-                                        pos.x, pos.y);
+      data_device_update_position (drag_grab, &pos);
     }
 
   input = meta_wayland_seat_get_input (seat);
@@ -1042,7 +1099,6 @@ data_device_set_selection (struct wl_client   *client,
                            uint32_t            serial)
 {
   MetaWaylandDataDevice *data_device = wl_resource_get_user_data (resource);
-  MetaWaylandSeat *seat = wl_container_of (data_device, seat, data_device);
   MetaWaylandDataSource *source;
 
   if (source_resource)
@@ -1191,6 +1247,8 @@ meta_wayland_data_device_manager_init (MetaWaylandCompositor *compositor)
 			META_WL_DATA_DEVICE_MANAGER_VERSION,
 			compositor, bind_manager) == NULL)
     g_error ("Could not create data_device");
+
+  meta_wayland_init_toplevel_drag (compositor);
 }
 
 void
@@ -1244,7 +1302,6 @@ void
 meta_wayland_data_device_set_focus (MetaWaylandDataDevice *data_device,
                                     MetaWaylandSurface    *surface)
 {
-  MetaWaylandSeat *seat = wl_container_of (data_device, seat, data_device);
   struct wl_client *focus_client = NULL;
   struct wl_resource *data_device_resource;
 
@@ -1286,3 +1343,12 @@ meta_wayland_data_device_unset_dnd_selection (MetaWaylandDataDevice *data_device
 {
   unset_selection_source (data_device, META_SELECTION_DND);
 }
+
+MetaWaylandToplevelDrag *
+meta_wayland_data_device_get_toplevel_drag (MetaWaylandDataDevice *data_device)
+{
+  if (!data_device->current_grab || !data_device->current_grab->drag_data_source)
+    return NULL;
+
+  return meta_wayland_data_source_get_toplevel_drag (data_device->current_grab->drag_data_source);
+}
