@@ -264,7 +264,7 @@ gtk_surface_titlebar_gesture (struct wl_client   *client,
       if (!window->can_grab)
         return;
 
-      if (META_WINDOW_MAXIMIZED (window))
+      if (meta_window_is_maximized (window))
         meta_window_unmaximize (window, META_MAXIMIZE_BOTH);
       else
         meta_window_maximize (window, META_MAXIMIZE_BOTH);
@@ -277,7 +277,7 @@ gtk_surface_titlebar_gesture (struct wl_client   *client,
       if (!window->can_grab)
         return;
 
-      if (META_WINDOW_MAXIMIZED_HORIZONTALLY (window))
+      if (meta_window_get_maximized (window) & META_MAXIMIZE_HORIZONTAL)
         meta_window_unmaximize (window, META_MAXIMIZE_HORIZONTAL);
       else
         meta_window_maximize (window, META_MAXIMIZE_HORIZONTAL);
@@ -290,7 +290,7 @@ gtk_surface_titlebar_gesture (struct wl_client   *client,
       if (!window->can_grab)
         return;
 
-      if (META_WINDOW_MAXIMIZED_VERTICALLY (window))
+      if (meta_window_get_maximized (window) & META_MAXIMIZE_VERTICAL)
         meta_window_unmaximize (window, META_MAXIMIZE_VERTICAL);
       else
         meta_window_maximize (window, META_MAXIMIZE_VERTICAL);
@@ -313,7 +313,7 @@ gtk_surface_titlebar_gesture (struct wl_client   *client,
       break;
 
     case G_DESKTOP_TITLEBAR_ACTION_MENU:
-      meta_window_show_menu (window, META_WINDOW_MENU_WM, x, y);
+      meta_window_show_menu (window, META_WINDOW_MENU_WM, (int) x, (int) y);
       break;
 
     default:
