@@ -256,10 +256,10 @@ meta_wayland_text_input_focus_delete_surrounding (ClutterInputFocus *focus,
   cursor = start + text_input->surrounding.cursor;
 
   before = g_utf8_offset_to_pointer (cursor, offset);
-  g_assert (before >= start);
+  g_return_if_fail (before >= start);
 
   after = g_utf8_offset_to_pointer (cursor, offset + len);
-  g_assert (after <= end);
+  g_return_if_fail (after <= end);
 
   before_length = cursor - before;
   after_length = after - cursor;
@@ -412,6 +412,27 @@ move_resources_for_client (struct wl_list *destination,
     }
 }
 
+static void
+reset_text_input_focus (MetaWaylandTextInput *text_input)
+{
+  if (wl_list_empty (&text_input->focus_resource_list))
+    {
+      ClutterInputFocus *focus = text_input->input_focus;
+      MetaBackend *backend = backend_from_text_input (text_input);
+      ClutterBackend *clutter_backend =
+        meta_backend_get_clutter_backend (backend);
+      ClutterInputMethod *input_method;
+
+      if (clutter_input_focus_is_focused (focus))
+        {
+          input_method = clutter_backend_get_input_method (clutter_backend);
+          clutter_input_focus_reset (focus);
+          meta_wayland_text_input_focus_flush_done (focus);
+          clutter_input_method_focus_out (input_method);
+        }
+    }
+}
+
 void
 meta_wayland_text_input_set_focus (MetaWaylandTextInput *text_input,
 				   MetaWaylandSurface   *surface)
@@ -425,18 +446,8 @@ meta_wayland_text_input_set_focus (MetaWaylandTextInput *text_input,
     {
       if (!wl_list_empty (&text_input->focus_resource_list))
         {
-          ClutterInputFocus *focus = text_input->input_focus;
-          ClutterInputMethod *input_method;
           struct wl_resource *resource;
 
-          if (clutter_input_focus_is_focused (focus))
-            {
-              input_method = clutter_backend_get_input_method (clutter_get_default_backend ());
-              clutter_input_focus_reset (focus);
-              meta_wayland_text_input_focus_flush_done (focus);
-              clutter_input_method_focus_out (input_method);
-            }
-
           wl_resource_for_each (resource, &text_input->focus_resource_list)
             {
               /* wayland-server doesn't allow sending a enter/leave events with
@@ -451,10 +462,19 @@ meta_wayland_text_input_set_focus (MetaWaylandTextInput *text_input,
 
           move_resources (&text_input->resource_list,
                           &text_input->focus_resource_list);
+
+          reset_text_input_focus (text_input);
         }
 
       wl_list_remove (&text_input->surface_listener.link);
       text_input->surface = NULL;
+      /* Wayland set_surrounding_text() does not support to set null string
+       * for applications with the non-supported surrounding text feature
+       * and reset the values here with focus changes.
+       */
+      g_clear_pointer (&text_input->surrounding.text, g_free);
+      text_input->surrounding.cursor = 0;
+      text_input->surrounding.anchor = 0;
     }
 
   if (surface && surface->resource)
@@ -496,8 +516,6 @@ meta_wayland_text_input_set_focus (MetaWaylandTextInput *text_input,
 
       if (!wl_list_empty (&text_input->focus_resource_list))
         {
-          struct wl_resource *resource;
-
           wl_resource_for_each (resource, &text_input->focus_resource_list)
             {
               /* wayland-server doesn't allow sending a enter/leave events with
@@ -520,6 +538,7 @@ text_input_destructor (struct wl_resource *resource)
 
   g_hash_table_remove (text_input->resource_serials, resource);
   wl_list_remove (wl_resource_get_link (resource));
+  reset_text_input_focus (text_input);
 
   /* We unset the magic resource and allow setting it again in case the magic
    * client goes away. This is so that the magic client can recover from crashes.
@@ -626,6 +645,7 @@ text_input_set_surrounding_text (struct wl_client   *client,
                                  int32_t             anchor)
 {
   MetaWaylandTextInput *text_input = wl_resource_get_user_data (resource);
+  size_t text_len = strlen (text);
 
   if (g_strcmp0 (text, "de9841d2-6324-4681-947b-c7902607e387.magictext-req") == 0)
     {
@@ -673,6 +693,13 @@ text_input_set_surrounding_text (struct wl_client   *client,
   if (!client_matches_focus (text_input, client))
     return;
 
+  if (cursor < 0 || anchor < 0 || cursor > text_len || anchor > text_len)
+    {
+      g_warning ("Client sent invalid surrounding text (text_len=%lu, cursor=%d, "
+                 "anchor=%d), ignoring", text_len, cursor, anchor);
+      return;
+    }
+
   g_free (text_input->pending_surrounding.text);
   text_input->pending_surrounding.text = g_strdup (text);
   text_input->pending_surrounding.cursor = cursor;
@@ -811,6 +838,9 @@ text_input_commit_state (struct wl_client   *client,
   MetaWaylandTextInput *text_input = wl_resource_get_user_data (resource);
   ClutterInputFocus *focus = text_input->input_focus;
   gboolean enable_panel = FALSE;
+  MetaBackend *backend = backend_from_text_input (text_input);
+  ClutterBackend *clutter_backend =
+    meta_backend_get_clutter_backend (backend);
   ClutterInputMethod *input_method;
 
   increment_serial (text_input, resource);
@@ -818,7 +848,7 @@ text_input_commit_state (struct wl_client   *client,
   if (!client_matches_focus (text_input, client))
     return;
 
-  input_method = clutter_backend_get_input_method (clutter_get_default_backend ());
+  input_method = clutter_backend_get_input_method (clutter_backend);
 
   if (input_method &&
       text_input->pending_state & META_WAYLAND_PENDING_STATE_ENABLED)
