@@ -39,6 +39,8 @@
 #include "wayland/meta-wayland-actor-surface.h"
 #include "wayland/meta-wayland-private.h"
 #include "wayland/meta-wayland-surface-private.h"
+#include "wayland/meta-wayland-toplevel-drag.h"
+#include "wayland/meta-wayland-transaction.h"
 #include "wayland/meta-wayland-window-configuration.h"
 #include "wayland/meta-wayland-xdg-shell.h"
 
@@ -238,7 +240,6 @@ meta_window_wayland_grab_op_ended (MetaWindow *window,
 
 static void
 meta_window_wayland_move_resize_internal (MetaWindow                *window,
-                                          MetaGravity                gravity,
                                           MtkRectangle               unconstrained_rect,
                                           MtkRectangle               constrained_rect,
                                           MtkRectangle               temporary_rect,
@@ -250,18 +251,20 @@ meta_window_wayland_move_resize_internal (MetaWindow                *window,
   MetaWindowWayland *wl_window = META_WINDOW_WAYLAND (window);
   gboolean can_move_now = FALSE;
   MtkRectangle configured_rect;
+  MtkRectangle frame_rect;
+  MetaGravity gravity;
   int geometry_scale;
   int new_x;
   int new_y;
   int new_buffer_x;
   int new_buffer_y;
 
-  g_assert (window->frame == NULL);
-
   /* don't do anything if we're dropping the window, see #751847 */
   if (window->unmanaging)
     return;
 
+  gravity = meta_window_get_gravity (window);
+
   configured_rect.x = constrained_rect.x;
   configured_rect.y = constrained_rect.y;
 
@@ -270,6 +273,7 @@ meta_window_wayland_move_resize_internal (MetaWindow                *window,
    * coordinate space so that we can have a scale independent size to pass
    * to the Wayland surface. */
   geometry_scale = meta_window_wayland_get_geometry_scale (window);
+  frame_rect = meta_window_config_get_rect (window->config);
 
   configured_rect.width = constrained_rect.width;
   configured_rect.height = constrained_rect.height;
@@ -303,20 +307,21 @@ meta_window_wayland_move_resize_internal (MetaWindow                *window,
           new_width = unconstrained_rect.width;
           new_height = unconstrained_rect.height;
         }
-      if (window->rect.width != new_width ||
-          window->rect.height != new_height)
+      if (frame_rect.width != new_width ||
+          frame_rect.height != new_height)
         {
           *result |= META_MOVE_RESIZE_RESULT_RESIZED;
-          window->rect.width = new_width;
-          window->rect.height = new_height;
+          meta_window_config_set_size (window->config,
+                                       new_width, new_height);
         }
 
+      frame_rect = meta_window_config_get_rect (window->config);
       window->buffer_rect.width =
-        window->rect.width +
+        frame_rect.width +
         window->custom_frame_extents.left +
         window->custom_frame_extents.right;
       window->buffer_rect.height =
-        window->rect.height +
+        frame_rect.height +
         window->custom_frame_extents.top +
         window->custom_frame_extents.bottom;
 
@@ -342,8 +347,8 @@ meta_window_wayland_move_resize_internal (MetaWindow                *window,
                 if (flags & META_MOVE_RESIZE_PLACEMENT_CHANGED ||
                     rel_x != wl_window->last_sent_rel_x ||
                     rel_y != wl_window->last_sent_rel_y ||
-                    constrained_rect.width != window->rect.width ||
-                    constrained_rect.height != window->rect.height)
+                    constrained_rect.width != frame_rect.width ||
+                    constrained_rect.height != frame_rect.height)
                   {
                     MetaWaylandWindowConfiguration *configuration;
 
@@ -377,8 +382,8 @@ meta_window_wayland_move_resize_internal (MetaWindow                *window,
               break;
             }
         }
-      else if (constrained_rect.width != window->rect.width ||
-               constrained_rect.height != window->rect.height ||
+      else if (constrained_rect.width != frame_rect.width ||
+               constrained_rect.height != frame_rect.height ||
                flags & META_MOVE_RESIZE_STATE_CHANGED)
         {
           MetaWaylandWindowConfiguration *configuration;
@@ -386,7 +391,7 @@ meta_window_wayland_move_resize_internal (MetaWindow                *window,
           int bounds_height;
 
           if (!meta_wayland_surface_get_buffer (wl_window->surface) &&
-              !META_WINDOW_MAXIMIZED (window) &&
+              !meta_window_is_maximized (window) &&
               window->tile_mode == META_TILE_NONE &&
               !meta_window_is_fullscreen (window))
             return;
@@ -434,11 +439,10 @@ meta_window_wayland_move_resize_internal (MetaWindow                *window,
         !!(flags & META_MOVE_RESIZE_STATE_CHANGED);
     }
 
-  if (new_x != window->rect.x || new_y != window->rect.y)
+  if (new_x != frame_rect.x || new_y != frame_rect.y)
     {
       *result |= META_MOVE_RESIZE_RESULT_MOVED;
-      window->rect.x = new_x;
-      window->rect.y = new_y;
+      meta_window_config_set_position (window->config, new_x, new_y);
     }
 
   if (window->placement.rule &&
@@ -471,14 +475,24 @@ scale_size (int  *width,
 {
   if (*width < G_MAXINT)
     {
-      float new_width = (*width * scale);
-      *width = (int) MIN (new_width, G_MAXINT);
+      float new_width;
+
+      new_width = (*width * scale);
+      if (new_width > G_MAXINT)
+        *width = G_MAXINT;
+      else
+        *width = (int) new_width;
     }
 
   if (*height < G_MAXINT)
     {
-      float new_height = (*height * scale);
-      *height = (int) MIN (new_height, G_MAXINT);
+      float new_height;
+
+      new_height = (*height * scale);
+      if (new_height > G_MAXINT)
+        *height = G_MAXINT;
+      else
+        *height = (int) new_height;
     }
 }
 
@@ -505,7 +519,7 @@ meta_window_wayland_update_main_monitor (MetaWindow                   *window,
   MetaLogicalMonitor *scaled_new;
   float from_scale, to_scale;
   float scale;
-  MtkRectangle rect;
+  MtkRectangle frame_rect;
 
   from = window->monitor;
 
@@ -521,7 +535,8 @@ g_return_if_fail (toplevel_window != NULL);
       return;
     }
 
-  if (window->rect.width == 0 || window->rect.height == 0)
+  frame_rect = meta_window_config_get_rect (window->config);
+  if (frame_rect.width == 0 || frame_rect.height == 0 || !window->placed)
     {
       window->monitor = meta_window_find_monitor_from_id (window);
       return;
@@ -568,10 +583,9 @@ g_return_if_fail (toplevel_window != NULL);
    * changes the main monitor, wait until both the current and the new scale
    * will result in the same main monitor. */
   scale = to_scale / from_scale;
-  rect = window->rect;
-  scale_rect_size (&rect, scale);
+  scale_rect_size (&frame_rect, scale);
   scaled_new =
-    meta_monitor_manager_get_logical_monitor_from_rect (monitor_manager, &rect);
+    meta_monitor_manager_get_logical_monitor_from_rect (monitor_manager, &frame_rect);
   if (to != scaled_new)
     return;
 
@@ -587,6 +601,7 @@ meta_window_wayland_main_monitor_changed (MetaWindow               *window,
   int geometry_scale;
   float scale_factor;
   MetaWaylandSurface *surface;
+  MtkRectangle frame_rect;
 
   if (!window->monitor)
     return;
@@ -610,7 +625,8 @@ meta_window_wayland_main_monitor_changed (MetaWindow               *window,
   scale_factor = (float) old_geometry_scale / geometry_scale;
 
   /* Window size. */
-  scale_rect_size (&window->rect, scale_factor);
+  frame_rect = meta_window_config_get_rect (window->config);
+  scale_rect_size (&frame_rect, scale_factor);
   scale_rect_size (&window->unconstrained_rect, scale_factor);
   scale_rect_size (&window->saved_rect, scale_factor);
   scale_size (&window->size_hints.min_width, &window->size_hints.min_height, scale_factor);
@@ -630,9 +646,9 @@ meta_window_wayland_main_monitor_changed (MetaWindow               *window,
   /* Buffer rect. */
   scale_rect_size (&window->buffer_rect, scale_factor);
   window->buffer_rect.x =
-    window->rect.x - window->custom_frame_extents.left;
+    frame_rect.x - window->custom_frame_extents.left;
   window->buffer_rect.y =
-    window->rect.y - window->custom_frame_extents.top;
+    frame_rect.y - window->custom_frame_extents.top;
 
   meta_compositor_sync_window_geometry (window->display->compositor,
                                         window,
@@ -709,6 +725,21 @@ on_window_shown (MetaWindow *window)
     meta_compositor_sync_updates_frozen (window->display->compositor, window);
 }
 
+static MetaWaylandToplevelDrag *
+get_toplevel_drag (MetaWindow *window)
+{
+  MetaWaylandToplevelDrag *toplevel_drag;
+  MetaWindowWayland *wl_window = META_WINDOW_WAYLAND (window);
+  MetaDisplay *display = meta_window_get_display (window);
+  MetaContext *context = meta_display_get_context (display);
+  MetaWaylandCompositor *compositor = meta_context_get_wayland_compositor (context);
+
+  toplevel_drag = meta_wayland_data_device_get_toplevel_drag (&compositor->seat->data_device);
+  if (!toplevel_drag || toplevel_drag->dragged_surface != wl_window->surface)
+    return NULL;
+  return toplevel_drag;
+}
+
 static void
 meta_window_wayland_init (MetaWindowWayland *wl_window)
 {
@@ -801,22 +832,55 @@ meta_window_wayland_set_transient_for (MetaWindow *window,
   return TRUE;
 }
 
+static void
+meta_window_wayland_stage_to_protocol (MetaWindow *window,
+                                       int         stage_x,
+                                       int         stage_y,
+                                       int        *protocol_x,
+                                       int        *protocol_y)
+{
+  if (protocol_x)
+    *protocol_x = stage_x;
+  if (protocol_y)
+    *protocol_y = stage_y;
+}
+
+static void
+meta_window_wayland_protocol_to_stage (MetaWindow          *window,
+                                       int                  protocol_x,
+                                       int                  protocol_y,
+                                       int                 *stage_x,
+                                       int                 *stage_y,
+                                       MtkRoundingStrategy  rounding_strategy)
+{
+  if (stage_x)
+    *stage_x = protocol_x;
+  if (stage_y)
+    *stage_y = protocol_y;
+}
+
+static MetaGravity
+meta_window_wayland_get_gravity (MetaWindow *window)
+{
+  MetaWindowWayland *wl_window = META_WINDOW_WAYLAND (window);
+  MetaWaylandSurface *surface = wl_window->surface;
+  MetaWaylandToplevelDrag *toplevel_drag;
+
+  /* Force nortwest gravity on toplevel drags */
+  toplevel_drag = get_toplevel_drag (window);
+
+  if (toplevel_drag && surface == toplevel_drag->dragged_surface)
+    return META_GRAVITY_NORTH_WEST;
+
+  return META_WINDOW_CLASS (meta_window_wayland_parent_class)->get_gravity (window);
+}
+
 static MetaStackLayer
 meta_window_wayland_calculate_layer (MetaWindow *window)
 {
   return meta_window_get_default_layer (window);
 }
 
-static void
-meta_window_wayland_map (MetaWindow *window)
-{
-}
-
-static void
-meta_window_wayland_unmap (MetaWindow *window)
-{
-}
-
 static void
 meta_window_wayland_constructed (GObject *object)
 {
@@ -825,10 +889,6 @@ meta_window_wayland_constructed (GObject *object)
   window->client_type = META_WINDOW_CLIENT_TYPE_WAYLAND;
 
   window->override_redirect = FALSE;
-  window->rect.x = 0;
-  window->rect.y = 0;
-  window->rect.width = 0;
-  window->rect.height = 0;
   /* size_hints are the "request" */
   window->size_hints.x = 0;
   window->size_hints.y = 0;
@@ -842,6 +902,8 @@ meta_window_wayland_constructed (GObject *object)
   window->decorated = FALSE;
   window->hidden = TRUE;
 
+  window->config = meta_window_config_new ();
+
   G_OBJECT_CLASS (meta_window_wayland_parent_class)->constructed (object);
 }
 
@@ -926,11 +988,12 @@ meta_window_wayland_class_init (MetaWindowWaylandClass *klass)
   window_class->can_ping = meta_window_wayland_can_ping;
   window_class->are_updates_frozen = meta_window_wayland_are_updates_frozen;
   window_class->calculate_layer = meta_window_wayland_calculate_layer;
-  window_class->map = meta_window_wayland_map;
-  window_class->unmap = meta_window_wayland_unmap;
   window_class->is_focus_async = meta_window_wayland_is_focus_async;
   window_class->get_wayland_surface = meta_window_wayland_get_wayland_surface;
   window_class->set_transient_for = meta_window_wayland_set_transient_for;
+  window_class->stage_to_protocol = meta_window_wayland_stage_to_protocol;
+  window_class->protocol_to_stage = meta_window_wayland_protocol_to_stage;
+  window_class->get_gravity = meta_window_wayland_get_gravity;
 
   obj_props[PROP_SURFACE] =
     g_param_spec_object ("surface", NULL, NULL,
@@ -1069,9 +1132,7 @@ meta_window_wayland_is_resize (MetaWindowWayland *wl_window,
   else
     {
       MetaWindow *window = META_WINDOW (wl_window);
-
-      old_width = window->rect.width;
-      old_height = window->rect.height;
+      meta_window_config_get_size (window->config, &old_width, &old_height);
     }
 
   return !wl_window->has_last_sent_configuration ||
@@ -1086,7 +1147,7 @@ meta_window_wayland_get_geometry_scale (MetaWindow *window)
     return 1;
 
   if (meta_window_is_alien (window))
-    return meta_logical_monitor_get_scale (window->monitor);
+    return (int) ceilf (meta_logical_monitor_get_scale (window->monitor));
 
   return 1;
 }
@@ -1146,13 +1207,15 @@ meta_window_wayland_finish_move_resize (MetaWindow              *window,
   MetaWaylandSurface *surface = wl_window->surface;
   int dx, dy;
   int geometry_scale;
-  MetaGravity gravity;
   MtkRectangle rect;
   MetaMoveResizeFlags flags;
   MetaWaylandWindowConfiguration *acked_configuration;
   gboolean is_window_being_resized;
   gboolean is_client_resize;
   MetaWindowDrag *window_drag;
+  MtkRectangle frame_rect;
+  MetaWindowActor *window_actor;
+  MetaWaylandToplevelDrag *toplevel_drag;
 
   /* new_geom is in the logical pixel coordinate space, but MetaWindow wants its
    * rects to represent what in turn will end up on the stage, i.e. we need to
@@ -1200,9 +1263,10 @@ meta_window_wayland_finish_move_resize (MetaWindow              *window,
      meta_grab_op_is_resizing (meta_window_drag_get_grab_op (window_drag)) &&
      meta_window_drag_get_window (window_drag) == window);
 
+  frame_rect = meta_window_config_get_rect (window->config);
   rect = (MtkRectangle) {
-    .x = window->rect.x,
-    .y = window->rect.y,
+    .x = frame_rect.x,
+    .y = frame_rect.y,
     .width = new_geom.width,
     .height = new_geom.height
   };
@@ -1214,10 +1278,12 @@ meta_window_wayland_finish_move_resize (MetaWindow              *window,
           if (window->placement.rule)
             {
               MetaWindow *parent;
+              MtkRectangle parent_rect;
 
               parent = meta_window_get_transient_for (window);
-              rect.x = parent->rect.x + acked_configuration->rel_x;
-              rect.y = parent->rect.y + acked_configuration->rel_y;
+              parent_rect = meta_window_config_get_rect (parent->config);
+              rect.x = parent_rect.x + acked_configuration->rel_x;
+              rect.y = parent_rect.y + acked_configuration->rel_y;
             }
           else
             {
@@ -1234,10 +1300,18 @@ meta_window_wayland_finish_move_resize (MetaWindow              *window,
         calculate_position (acked_configuration, &new_geom, &rect);
     }
 
+  toplevel_drag = get_toplevel_drag (window);
+  if (toplevel_drag && !is_window_being_resized && !window->mapped &&
+      rect.width > 0 && rect.height > 0)
+    {
+      meta_wayland_toplevel_drag_calc_origin_for_dragged_window (toplevel_drag,
+                                                                 &rect);
+    }
+
   rect.x += dx;
   rect.y += dy;
 
-  if (rect.x != window->rect.x || rect.y != window->rect.y)
+  if (rect.x != frame_rect.x || rect.y != frame_rect.y)
     flags |= META_MOVE_RESIZE_MOVE_ACTION;
 
   if (wl_window->has_pending_state_change && acked_configuration)
@@ -1246,7 +1320,7 @@ meta_window_wayland_finish_move_resize (MetaWindow              *window,
       wl_window->has_pending_state_change = FALSE;
     }
 
-  if (rect.width != window->rect.width || rect.height != window->rect.height)
+  if (rect.width != frame_rect.width || rect.height != frame_rect.height)
     {
       flags |= META_MOVE_RESIZE_RESIZE_ACTION;
 
@@ -1261,12 +1335,14 @@ meta_window_wayland_finish_move_resize (MetaWindow              *window,
                    meta_wayland_window_configuration_free);
   wl_window->last_acked_configuration = g_steal_pointer (&acked_configuration);
 
-  if (window_drag &&
-      meta_window_drag_get_window (window_drag) == window)
-    gravity = meta_resize_gravity_from_grab_op (meta_window_drag_get_grab_op (window_drag));
-  else
-    gravity = META_GRAVITY_STATIC;
-  meta_window_move_resize_internal (window, flags, gravity, rect);
+  /* Force unconstrained move when running toplevel drags */
+  if (toplevel_drag && surface == toplevel_drag->dragged_surface)
+    {
+      window_actor = meta_window_actor_from_window (window);
+      meta_window_actor_set_tied_to_drag (window_actor, TRUE);
+    }
+
+  meta_window_move_resize (window, flags, rect);
 }
 
 void
@@ -1281,19 +1357,19 @@ meta_window_place_with_placement_rule (MetaWindow        *window,
   window->placement.rule = g_new0 (MetaPlacementRule, 1);
   *window->placement.rule = *placement_rule;
 
-  window->unconstrained_rect.x = window->rect.x;
-  window->unconstrained_rect.y = window->rect.y;
+  meta_window_config_get_position (window->config,
+                                   &window->unconstrained_rect.x,
+                                   &window->unconstrained_rect.y);
   window->unconstrained_rect.width = placement_rule->width;
   window->unconstrained_rect.height = placement_rule->height;
 
   window->calc_placement = first_placement;
-  meta_window_move_resize_internal (window,
-                                    (META_MOVE_RESIZE_MOVE_ACTION |
-                                     META_MOVE_RESIZE_RESIZE_ACTION |
-                                     META_MOVE_RESIZE_PLACEMENT_CHANGED |
-                                     META_MOVE_RESIZE_CONSTRAIN),
-                                    META_GRAVITY_NORTH_WEST,
-                                    window->unconstrained_rect);
+  meta_window_move_resize (window,
+                           (META_MOVE_RESIZE_MOVE_ACTION |
+                            META_MOVE_RESIZE_RESIZE_ACTION |
+                            META_MOVE_RESIZE_PLACEMENT_CHANGED |
+                            META_MOVE_RESIZE_CONSTRAIN),
+                           window->unconstrained_rect);
   window->calc_placement = FALSE;
 }
 
@@ -1325,7 +1401,7 @@ meta_window_wayland_set_min_size (MetaWindow *window,
       return;
     }
 
-  scale = 1.0 / (float) meta_window_wayland_get_geometry_scale (window);
+  scale = 1.0f / (float) meta_window_wayland_get_geometry_scale (window);
   scale_size (&width, &height, scale);
 
   new_width = width + (window->custom_frame_extents.left +
@@ -1359,7 +1435,7 @@ meta_window_wayland_set_max_size (MetaWindow *window,
       return;
     }
 
-  scale = 1.0 / (float) meta_window_wayland_get_geometry_scale (window);
+  scale = 1.0f / (float) meta_window_wayland_get_geometry_scale (window);
   scale_size (&width, &height, scale);
 
   new_width = width + (window->custom_frame_extents.left +
@@ -1446,3 +1522,19 @@ meta_window_wayland_is_acked_fullscreen (MetaWindowWayland *wl_window)
   return (wl_window->last_acked_configuration &&
           wl_window->last_acked_configuration->is_fullscreen);
 }
+
+gboolean
+meta_window_wayland_get_pending_serial (MetaWindowWayland *wl_window,
+                                        uint32_t          *serial)
+{
+  if (wl_window->pending_configurations)
+    {
+      MetaWaylandWindowConfiguration *configuration =
+        wl_window->pending_configurations->data;
+
+      *serial = configuration->serial;
+      return TRUE;
+    }
+
+  return FALSE;
+}
